Require Import List ListSet Program.

Require Import CSL.Syntax.
Require Import CSL.Definitions.
Require Import CSL.HList.
Require Import CSL.Analysis.Lattice.

Set Implicit Arguments.

(* Approximation of powerset domain of values *)
Inductive abstract_set τ : Type :=
| FullSet : abstract_set τ
| ActualSet : set (tyDenote τ) -> abstract_set τ.

(* Abstract value for use in analysis *)
Inductive abstract_value τ : Type :=
| AnyValue : abstract_value τ
| ActualValue : tyDenote τ -> abstract_value τ.

Arguments FullSet {τ}.
Arguments ActualSet [τ].
Arguments AnyValue {τ}.
Arguments ActualValue [τ].

Definition union_abstract_set τ (s1 s2 : abstract_set τ) : abstract_set τ :=
  match s1, s2 with
  | FullSet, _ => FullSet
  | _, FullSet => FullSet
  | ActualSet s1', ActualSet s2' => ActualSet (set_union (tyDenote_dec_eq τ) s1' s2')
  end.

Definition intersect_abstract_set τ (s1 s2 : abstract_set τ) : abstract_set τ :=
  match s1, s2 with
  | FullSet, _ => s2
  | _, FullSet => s1
  | ActualSet s1', ActualSet s2' => ActualSet (set_inter (tyDenote_dec_eq τ) s1' s2')
  end.

Lemma set_inter_subset:
  forall A dec (s1 s2 : set A),
    (forall e, set_In e s1 -> set_In e s2) ->
    set_inter dec s1 s2 = s1.
Proof.
  induction s1; intros; simpl; [reflexivity | ].
  remember (set_mem dec a s2).
  destruct b.
  rewrite IHs1; auto. intros; apply H. simpl; auto.
  simpl in *.
  specialize H with a.
  symmetry in Heqb. apply set_mem_complete1 in Heqb.
  intuition.
Qed.

Lemma set_inter_id:
  forall A dec (s : set A),
    set_inter dec s s = s.
Proof.
  induction s; simpl; intuition.
  destruct (dec a a); [ | contradiction].
  rewrite set_inter_subset; [reflexivity | intros; simpl; auto].
Qed.

Lemma intersect_abstract_set_id:
  forall τ (s : abstract_set τ),
    intersect_abstract_set s s = s.
Proof.
  induction s; simpl; [ | rewrite set_inter_id ]; reflexivity.
Qed.

Definition value_map Δ := hlist abstract_set Δ.

Fixpoint empty_map Δ : value_map Δ :=
  match Δ with
  | nil => HNil
  | τ :: Δ' => HCons (ActualSet (empty_set (tyDenote τ))) (empty_map Δ')
  end.

Fixpoint full_map Δ : value_map Δ :=
  match Δ with
  | nil => HNil
  | τ :: Δ' => HCons FullSet (full_map Δ')
  end.

Fixpoint intersect_all Δ : value_map Δ -> value_map Δ -> value_map Δ :=
  match Δ return value_map Δ -> value_map Δ -> value_map Δ with
  | nil => fun _ _ => HNil
  | _ :: _ => fun m1 m2 => HCons (intersect_abstract_set (hhead m1) (hhead m2)) (intersect_all (htail m1) (htail m2))
  end.

Fixpoint union_all Δ : value_map Δ -> value_map Δ -> value_map Δ :=
  match Δ return value_map Δ -> value_map Δ -> value_map Δ with
  | nil => fun _ _ => HNil
  | _ :: _ => fun m1 m2 => HCons (union_abstract_set (hhead m1) (hhead m2)) (union_all (htail m1) (htail m2))
  end.

(* Mapping of new variables to unknown values for input to predicate analysis *)
Definition new_var_value_map (Δ : list ty) (m : value_map Δ) : value_map (new_var Δ) :=
    @HCons ty abstract_set Agent _ FullSet
           (@HCons ty _ Agent _ FullSet
                   (@HCons ty _ Resource _ FullSet
                           (@HCons ty _ _ Δ FullSet m))).

Definition singleton A Aeq_dec (v : A) := set_add Aeq_dec v (empty_set A).

Definition abstract_set_In τ (v : tyDenote τ) (s : abstract_set τ) : Prop :=
  match s with
  | FullSet => True
  | ActualSet s' => set_In v s'
  end.

Definition abstract_set_Incl τ (s s' : abstract_set τ) : Prop :=
  match s, s' with
  | FullSet, FullSet => True
  | ActualSet s1, FullSet => True
  | ActualSet s1, ActualSet s2 => forall v, set_In v s1 -> set_In v s2
  | FullSet, ActualSet s2 => False
  end.

(* An environment agrees with the value map (overapproximation) *)
Definition value_map_correct Δ (δ : env Δ) (m : value_map Δ) :=
  forall τ (v : member τ Δ), abstract_set_In (hget δ v) (hget m v).

(* A value map is more precise that another one *)
Definition value_map_Incl Δ (m m' : value_map Δ) :=
  forall τ (x : member τ Δ), abstract_set_Incl (hget m x) (hget m' x).

Hint Resolve set_inter_intro set_union_intro : csl.

Lemma abstract_set_inter_intro : forall τ (a a0 : abstract_set τ) (b : tyDenote τ),
    abstract_set_In b a /\ abstract_set_In b a0 -> abstract_set_In b (intersect_abstract_set a a0).
Proof.
  unfold abstract_set_In; intros;
    destruct H; destruct a, a0; simpl; auto with csl.
Qed.

Lemma abstract_set_union_intro : forall τ (a a0 : abstract_set τ) (b : tyDenote τ),
    abstract_set_In b a \/ abstract_set_In b a0 -> abstract_set_In b (union_abstract_set a a0).
Proof.
  unfold abstract_set_In; intros;
    destruct H; destruct a, a0; simpl; auto with csl.
Qed.

Lemma hget_hput_independent_type:
  forall Δ τ τ0 (m : value_map Δ) (v : member τ Δ) (v' : member τ0 Δ) (s : abstract_set τ),
  τ <> τ0 ->
  hget (hput s m v) v' = hget m v'.
Proof.
  intros; dependent induction m; dependent destruction v'; dependent destruction v; try contradiction; auto.
  simpl; apply IHm; auto.
Qed.

Lemma hget_hput_independent_var:
  forall Δ τ (m : value_map Δ) (v v': member τ Δ) (s : abstract_set τ),
  v <> v' ->
  hget (hput s m v) v' = hget m v'.
Proof.
  intros; dependent induction m; dependent destruction v'; dependent destruction v; try contradiction; auto.
  simpl; apply IHm; intro; elim H; subst; auto.
Qed.

Lemma abstract_set_Incl_abstract_set_In:
  forall τ (s : abstract_set τ) s0,
     abstract_set_Incl s s0 ->
    (forall x, abstract_set_In x s -> abstract_set_In x s0).
Proof.
  induction s0; try dependent destruction s; simpl; intuition.
Qed.

Lemma abstract_set_Incl_refl:
  forall τ (s : abstract_set τ),
    abstract_set_Incl s s.
Proof.
  dependent destruction s; simpl; auto.
Qed.

Lemma abstract_set_Incl_trans:
  forall τ (s1 s2 s3: abstract_set τ),
    abstract_set_Incl s1 s2 ->
    abstract_set_Incl s2 s3 ->
    abstract_set_Incl s1 s3.
Proof.
  dependent destruction s1; dependent destruction s2; dependent destruction s3; intros; simpl in *; auto; contradiction.
Qed.

Lemma abstract_set_Incl_HCons:
  forall Δ (m1 m2 : value_map Δ) τ (a a0 : abstract_set τ),
    value_map_Incl m1 m2 ->
    abstract_set_Incl a a0 ->
    value_map_Incl (HCons a m1) (HCons a m2).
Proof.
  unfold value_map_Incl; intros; subst; dependent destruction x; simpl; auto.
  apply abstract_set_Incl_refl.
Qed.


(* Nasty proofs due to the "nice" definition of value_map_correct *)
Lemma value_map_correct_intersect : forall Δ (δ : env Δ) (m1 m2 : value_map Δ),
    value_map_correct δ m1 /\ value_map_correct δ m2 -> value_map_correct δ (intersect_all m1 m2).
Proof.
  dependent induction δ;
    unfold value_map_correct, abstract_set_In in *; intros.
  * inversion v.
  * destruct H; dependent destruction m1;
      dependent destruction m2; specialize (IHδ m1 m2).
    dependent destruction v.
    - specialize (H x HFirst); specialize (H0 x HFirst).
      destruct a, a0; simpl; auto with csl.
    - apply IHδ.
      split; intros;
        specialize (H τ0 (HNext v0)); specialize (H0 τ0 (HNext v0)); auto.
Qed.

Lemma value_map_correct_union : forall Δ (δ : env Δ) (m1 m2 : value_map Δ),
    value_map_correct δ m1 \/ value_map_correct δ m2 -> value_map_correct δ (union_all m1 m2).
Proof.
  dependent induction δ;
    unfold value_map_correct; intros.
  - inversion v.
  - let case1 := specialize (H x HFirst); simpl in *; apply abstract_set_union_intro; auto
     in dependent destruction m1; dependent destruction m2;
          specialize (IHδ m1 m2); intros; destruct H;
            unfold value_map_correct in *;
            dependent destruction v; simpl;
              [ case1
              | apply IHδ;
                intros;
                left;
                intros;
                specialize (H τ0 (HNext v0));
                auto
              | case1
              | apply IHδ;
                intros;
                right;
                intros;
                specialize (H τ0 (HNext v0));
                auto
              ].
Qed.

Lemma value_map_trivial : forall Δ (δ : env Δ),
    value_map_correct δ (full_map Δ).
Proof.
  dependent induction δ.
  - unfold value_map_correct, abstract_set_In in *; intros; inversion v.
  - dependent destruction v; simpl; auto.
Qed.

(* Someone should make a proof principle for value_map_correct and friends *)
Hint Resolve set_inter_intro : csl.

Lemma put_intersect_var_val : forall Δ (δ : env Δ) (m : value_map Δ) τ (a : tyDenote τ) (v : member τ Δ),
      value_map_correct δ m ->
      hget δ v = a ->
      value_map_correct δ (hput (intersect_abstract_set (hget m v) (ActualSet (singleton (tyDenote_dec_eq τ) a))) m v).
Proof.
  intros; subst; dependent induction v;
    dependent destruction m;
    dependent destruction δ;
    unfold value_map_correct in *; intros.

  * dependent destruction v.
    - destruct a; simpl; auto.
      apply set_inter_intro; simpl; auto.
      specialize (H τ HFirst); auto.
    - specialize (H τ0 (HNext v)); auto.

  * dependent destruction v0.
    - specialize (H x HFirst); auto.
    - apply IHv; intros; specialize (H τ1 (HNext v1)); auto.
Qed.

Lemma intersect_set_hget_hput:
  forall Δ τ  (m: value_map Δ) (v : member τ Δ) s1 s2 k,
    abstract_set_In k s1 -> abstract_set_In k s2 ->
    abstract_set_In k (hget (hput (intersect_abstract_set s1 s2) m v) v).
Proof.
  dependent induction m; dependent destruction v; simpl; intros;
    [ | rewrite hget_hput]; apply abstract_set_inter_intro; auto.
Qed.

Lemma put_intersect_var_var : forall Δ (δ : env Δ) (m : value_map Δ) τ (v0 v1 : member τ Δ),
    value_map_correct δ m ->
    hget δ v0 = hget δ v1 ->
    value_map_correct δ (hput (intersect_abstract_set (hget m v0) (hget m v1))
                              (hput (intersect_abstract_set (hget m v0) (hget m v1)) m v0) v1).
Proof.
  unfold value_map_correct; intros.
  destruct (ty_eq_dec τ τ0);
  [ subst | do 2 (rewrite hget_hput_independent_type; auto)];
  destruct (member_dec_eq v v1); destruct (member_dec_eq v v0); subst.
  - rewrite intersect_abstract_set_id. do 2 rewrite hput_hget_id. auto.
  - rewrite hget_hput; apply abstract_set_inter_intro; split; [ rewrite <- H0 | ]; auto.
  - rewrite hget_hput_independent_var; auto; apply intersect_set_hget_hput; [ | rewrite H0 ]; auto.
  - do 2 (rewrite hget_hput_independent_var; auto).
Qed.

Program Instance abstract_set_lattice τ : Lattice (abstract_set τ) :=
  {
    top := FullSet;
    bot := ActualSet (empty_set (tyDenote τ));
    Incl x y := abstract_set_Incl x y;
    join x y := union_abstract_set x y
  }.
Next Obligation.
  intros.
  destruct x; simpl; auto.
  intuition.
Defined.
Next Obligation.
  destruct x; simpl; auto.
Defined.
Next Obligation.
  destruct x; simpl; auto.
Defined.
Next Obligation.
  intros; unfold abstract_set_Incl in *; destruct x, y, z; intuition.
Defined.
Next Obligation.
  intros.
  unfold abstract_set_Incl in *;
    split; destruct x, y; simpl in *; intuition.
Defined.
Next Obligation.
  intros.
  unfold abstract_set_Incl.
  destruct x, x', y, y'; simpl in *; intuition.
  apply set_union_elim in H1; intuition.
Defined.

Program Instance abstract_set_setlattice : SetLattice abstract_set :=
  {
    In := abstract_set_In;
    lift {τ} v := ActualSet (singleton (tyDenote_dec_eq τ) v)
  }.
Next Obligation.
  intuition;
    destruct l; simpl in *; eauto.
  intros. intuition. subst; auto.
Defined.

Lemma abstract_intersect_FullSet: forall τ (s: abstract_set τ),
      intersect_abstract_set s FullSet = s.
Proof.
  destruct s; simpl; auto.
Qed.

Lemma abstract_set_Incl_Fullset:
  forall τ (s : abstract_set τ) , abstract_set_Incl s FullSet.
Proof.
  destruct s; simpl; auto.
Qed.
Hint Resolve abstract_set_Incl_Fullset : csl.

Lemma abstract_set_Incl_intersection:
  forall τ (a a' b b' : abstract_set τ),
    abstract_set_Incl a a' -> abstract_set_Incl b b'->
    abstract_set_Incl (intersect_abstract_set a b) (intersect_abstract_set a' b').
Proof.
  unfold abstract_set_Incl; dependent induction a; dependent destruction a'; dependent destruction b; dependent destruction b'; simpl in *; intros; auto;
    try contradiction;
  intuition;
  apply set_inter_elim in H1; intuition.
Qed.

Lemma abstract_set_Incl_union:
  forall τ (a a' b b' : abstract_set τ),
    abstract_set_Incl a a' -> abstract_set_Incl b b'->
    abstract_set_Incl (union_abstract_set a b) (union_abstract_set a' b').
Proof.
  unfold abstract_set_Incl; dependent induction a; dependent destruction a'; dependent destruction b; dependent destruction b'; simpl in *; intros; auto;
    try contradiction; intuition; apply set_union_elim in H1; intuition.
Qed.

Lemma abstract_set_Incl_union_l:
  forall τ (a a' b b' : abstract_set τ),
    abstract_set_Incl a a' -> abstract_set_Incl b b'->
    abstract_set_Incl a (union_abstract_set a' b').
Proof.
  unfold abstract_set_Incl; dependent induction a; dependent destruction a'; dependent destruction b; dependent destruction b'; simpl in *; intros; auto;
    try contradiction; intuition; apply set_union_intro; intuition.
Qed.

Lemma abstract_set_Incl_union_r:
  forall τ (a a' b b' : abstract_set τ),
    abstract_set_Incl a a' -> abstract_set_Incl b b'->
    abstract_set_Incl b (union_abstract_set a' b').
Proof.
  unfold abstract_set_Incl; dependent induction a; dependent destruction a'; dependent destruction b; dependent destruction b'; simpl in *; intros; auto;
    try contradiction; intuition; apply set_union_intro; intuition.
Qed.

Lemma abstract_set_Incl_union_simpl:
  forall τ (s1 s2 s3: abstract_set τ),
    abstract_set_Incl s1 s3 -> abstract_set_Incl s2 s3 -> abstract_set_Incl (union_abstract_set s1 s2) s3.
Proof.
  dependent destruction s1; dependent destruction s2; dependent destruction s3; intros; simpl in *; auto.
  intros. apply set_union_elim in H1; intuition.
Qed.

Lemma abstract_set_Incl_intersect_all:
  forall Δ (m m' n n' : value_map Δ),
    (forall τ (x : member τ Δ), abstract_set_Incl (hget m x) (hget m' x)) ->
    (forall τ (x : member τ Δ), abstract_set_Incl (hget n x) (hget n' x)) ->
    forall τ (x : member τ Δ), abstract_set_Incl (hget (intersect_all m n) x) (hget (intersect_all m' n') x).
Proof.
  dependent induction Δ; intros; [inversion x | ].
  dependent destruction x.
  - simpl; dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
    specialize H with a HFirst. simpl in H. specialize H0 with a HFirst; simpl in H0.
    apply abstract_set_Incl_intersection; auto.
  - apply IHΔ; intros.
    + dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
      specialize H with τ0 (HNext x); auto.
    + dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
      specialize H0 with τ0 (HNext x); auto.
Qed.

Lemma abstract_set_Incl_union_all:
  forall Δ (m m' n n' : value_map Δ),
    (forall τ (x : member τ Δ), abstract_set_Incl (hget m x) (hget m' x)) ->
    (forall τ (x : member τ Δ), abstract_set_Incl (hget n x) (hget n' x)) ->
    forall τ (x : member τ Δ), abstract_set_Incl (hget (union_all m n) x) (hget (union_all m' n') x).
Proof.
  dependent induction Δ; intros; [inversion x | ].
  dependent destruction x.
  - simpl. dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
    specialize H with a HFirst. simpl in H. specialize H0 with a HFirst; simpl in H0.
    apply abstract_set_Incl_union; auto.
  - apply IHΔ; intros.
    + dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
      specialize H with τ0 (HNext x); auto.
    + dependent destruction m; dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
      specialize H0 with τ0 (HNext x); auto.
Qed.

Lemma abstract_set_Incl_union_all_r:
  forall Δ (m' n n' : value_map Δ),
    (forall τ (x : member τ Δ), abstract_set_Incl (hget n x) (hget n' x)) ->
    forall τ (x : member τ Δ), abstract_set_Incl (hget n x) (hget (union_all m' n') x).
Proof.
  dependent induction Δ; intros; [inversion x | ].
  dependent destruction x.
  - simpl. dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
    specialize H with a HFirst. simpl in H. apply abstract_set_Incl_trans with (s2:= a2); auto.
    apply abstract_set_Incl_union_r with (a:=a0);  apply abstract_set_Incl_refl.
  - dependent destruction n; dependent destruction m'; dependent destruction n'; simpl.
    apply abstract_set_Incl_trans with (s2:= hget n' x1).
    specialize H with τ (HNext x1); simpl in H; auto.
    apply IHΔ; eauto; intros.
    apply abstract_set_Incl_refl.
Qed.

Lemma abstract_set_Incl_union_all_l:
  forall Δ (m m' n' : value_map Δ),
    (forall τ (x : member τ Δ), abstract_set_Incl (hget m x) (hget m' x)) ->
    forall τ (x : member τ Δ), abstract_set_Incl (hget m x) (hget (union_all m' n') x).
Proof.
  dependent induction Δ; intros; [inversion x | ].
  dependent destruction x.
  - simpl. dependent destruction m; dependent destruction m'; dependent destruction n'; simpl.
    specialize H with a HFirst. simpl in H. apply abstract_set_Incl_trans with (s2:= a1); auto.
    apply abstract_set_Incl_union_l with (b:=a2);  apply abstract_set_Incl_refl.
  - dependent destruction m; dependent destruction m'; dependent destruction n'; simpl.
    apply abstract_set_Incl_trans with (s2:= hget m' x1).
    specialize H with τ (HNext x1); simpl in H; auto.
    apply IHΔ; eauto; intros.
    apply abstract_set_Incl_refl.
Qed.

Lemma put_intersect_Incl_var_var : forall Δ (m m': value_map Δ) τ (v0 v1 : member τ Δ),
    value_map_Incl m m' ->
    value_map_Incl (hput (intersect_abstract_set (hget m v0) (hget m v1))
                         (hput (intersect_abstract_set (hget m v0) (hget m v1)) m v0) v1)
                   (hput (intersect_abstract_set (hget m' v0) (hget m' v1))
                         (hput (intersect_abstract_set (hget m' v0) (hget m' v1)) m' v0) v1).
Proof.
  unfold value_map_Incl; intros.
  destruct (ty_eq_dec τ τ0);
    [ subst | do 4 (rewrite hget_hput_independent_type; auto)].
  destruct (member_dec_eq x v1); destruct (member_dec_eq x v0); subst.
  - do 2 rewrite intersect_abstract_set_id. do 4 rewrite hput_hget_id. auto.
  - do 2 rewrite hget_hput. apply abstract_set_Incl_intersection; auto.
  - rewrite hget_hput_independent_var; auto;
    rewrite hget_hput_independent_var with (v:=v1); auto;
      do 2 rewrite hget_hput; apply abstract_set_Incl_intersection; auto.
  - do 4 (rewrite hget_hput_independent_var; auto).
Qed.

Lemma abstract_set_Incl_In:
  forall Δ τ τ' (m m' : value_map Δ) s1 s2 (v0: member τ Δ) (v1: member τ' Δ),
    value_map_Incl m m' ->
    abstract_set_Incl s1 s2 ->
    (forall x, abstract_set_In x (hget (hput s1 m v0) v1) ->
               abstract_set_In x (hget (hput s2 m' v0) v1)).
Proof.
  intros.
  destruct (ty_eq_dec τ τ').
  - subst; destruct (member_dec_eq v0 v1).
    + subst. rewrite hget_hput; rewrite hget_hput in H1.
      eapply abstract_set_Incl_abstract_set_In in H1; eauto.
    + unfold value_map_Incl in H;
        rewrite hget_hput_independent_var in H1; auto; rewrite hget_hput_independent_var; auto.
      specialize H with τ' v1;  eapply abstract_set_Incl_abstract_set_In in H1; eauto.
  - unfold value_map_Incl in H;
      rewrite hget_hput_independent_type in H1; auto; rewrite hget_hput_independent_type; auto.
    specialize H with τ' v1;  eapply abstract_set_Incl_abstract_set_In in H1; eauto.
Qed.

Lemma abstract_set_Incl_Incl:
  forall Δ τ τ' (m m' : value_map Δ) s1 s2 (v0: member τ Δ) (v1: member τ' Δ),
    value_map_Incl m m' ->
    abstract_set_Incl s1 s2 ->
    abstract_set_Incl (hget (hput s1 m v0) v1) (hget (hput s2 m' v0) v1).
Proof.
  intros.
  destruct (ty_eq_dec τ τ');
  [ | do 2 (rewrite hget_hput_independent_type; auto)];
  subst; destruct (member_dec_eq v0 v1);
    [ | do 2 (rewrite hget_hput_independent_var; auto)]; subst.
  do 2 rewrite hget_hput. assumption.
Qed.

Lemma put_intersect_Incl_var_val : forall Δ (m m': value_map Δ)  τ s (v0 v1 : member τ Δ),
    value_map_Incl m m' ->
    value_map_Incl
       (hput (intersect_abstract_set (hget m v0) s) m v1)
       (hput (intersect_abstract_set (hget m' v0) s) m' v1).
Proof.
  unfold value_map_Incl; intros.
  apply abstract_set_Incl_Incl;
    [ | apply abstract_set_Incl_intersection]; auto.
  apply abstract_set_Incl_refl.
Qed.
