Require Import Bool List Program String ListSet.
 
Require Import CSL.Syntax
        CSL.Definitions
        CSL.Exp
        CSL.HList
        CSL.Analysis.Generic
        CSL.Analysis.AbstractValues.

Set Implicit Arguments.

Definition unifyVarVar{τ Δ} (old : value_map Δ) (v v' : member τ Δ) : value_map Δ :=
  let newV := intersect_abstract_set (hget old v) (hget old v') in
  hput newV (hput newV old v) v'.

Hint Unfold unifyVarVar : csl.

Definition unifyVarLit{τ Δ} (old : value_map Δ) (v : member τ Δ) l : value_map Δ :=
  let oldV := hget old v in
    let newV := ActualSet (singleton (tyDenote_dec_eq _ ) l) in
    hput (intersect_abstract_set oldV newV) old v.

Hint Unfold unifyVarLit : csl.

(* Returning None => not possible to equate the two expressions *)
Definition equate τ Δ (old : value_map Δ) (e1 : exp Δ τ) (e2 : exp Δ τ) : option (value_map Δ) :=
  match e1 in exp _ τ, e2 in exp _ τ with
  (* equating variables means unifying them *)
  | @Var _ Agent v1, @Var _ Agent v2
  | @Var _ Resource v1, @Var _ Resource v2
  | @Var _ Timestamp v1, @Var _ Timestamp v2
  | @Var _ Bool v1, @Var _ Bool v2
  | @Var _ Scale v1, @Var _ Scale v2 => Some (unifyVarVar old v1 v2)
  (* equating literals means checking if they're, well, equal *)
  | @Lit _ Agent l1, @Lit _ Agent l2
  | @Lit _ Resource l1, @Lit _ Resource l2
  | @Lit _ Timestamp l1, @Lit _ Timestamp l2
  | @Lit _ Bool l1, @Lit _ Bool l2
  | @Lit _ Scale l1, @Lit _ Scale l2 => if (equalTy l1 l2) then Some old else None
  (* equating a variable and a literal means intersecting the values of the variable with the literal *)
  | @Lit _ Agent l, @Var _ Agent v
  | @Var _ Agent v, @Lit _ Agent l
  | @Lit _ Resource l, @Var _ Resource v
  | @Var _ Resource v, @Lit _ Resource l
  | @Lit _ Timestamp l, @Var _ Timestamp v
  | @Var _ Timestamp v, @Lit _ Timestamp l
  | @Lit _ Bool l, @Var _ Bool v
  | @Var _ Bool v, @Lit _ Bool l
  | @Lit _ Scale l, @Var _ Scale v
  | @Var _ Scale v, @Lit _ Scale l => Some (unifyVarLit old v l)
  | _, _ => Some old
end.

(*
 * This captures the notion of inferring values for variables.
 * We return None if it contains a contradiction
 * If we have an equality, we use the previous function to infer some
 * values (only for literal & variable combinations).
 *)
Fixpoint possibleValues Δ τ (old : value_map Δ) (e : exp Δ τ) : option (value_map Δ) :=
  match e with
  | @Lit _ Bool v => if v then Some old else None
  | @Unop _ Bool Bool u e => Some old 
  | Binop o e1 e2 =>
    match o in binop τ' τ'' _ return exp Δ τ' -> exp Δ τ'' -> option (value_map Δ) with
    | Eq _ => fun e1 e2 => equate old e1 e2
    | And => fun e1 e2 =>
               let pv1 := possibleValues old e1 in
               let pv2 := possibleValues old e2 in
               match pv1, pv2 with
               | Some m1, Some m2 => Some (intersect_all m1 m2)
               | _, _ => None
               end
    | Or => fun e1 e2 =>
              let pv1 := possibleValues old e1 in
              let pv2 := possibleValues old e2 in
              match pv1, pv2 with
              | Some m, None
              | None, Some m => Some m
              | Some m1, Some m2 => Some (union_all m1 m2)
              | _, _ => None
              end
    | _ => fun _ _ => Some old
    end e1 e2
  | _ => Some old
  end.

Hint Resolve value_map_correct_union value_map_correct_intersect value_map_trivial eqb_eq : csl.

Lemma noContradictions_equate: forall Δ τ (δ : env Δ) (m : value_map Δ) (e e' : exp Δ τ),
    binopDenote (Eq (tyToEquatable τ)) (expDenote e δ) (expDenote e' δ) = true ->
    exists m', equate m e e' = Some m'.
Proof.
  dependent induction e; dependent destruction e'; intros;
    try (try destruct τ; try destruct t3; eexists; simpl; reflexivity).
  - destruct t2; eexists; simpl; reflexivity.
  - simpl in H.
    destruct τ; eexists; simpl in *; subst; rewrite H; eauto.
  - destruct t2;  eexists; simpl; reflexivity.
Qed.

Lemma noContradictions : forall Δ (δ : env Δ) (m : value_map Δ) (e : exp Δ Bool),
    expDenote e δ = true -> exists m', possibleValues m e = Some m'.
Proof.
  dependent induction e; intros; simpl in *; subst; eauto.
  - dependent destruction b; simpl in *; eauto.
    + apply andb_true_iff in H as [H H'].
      apply IHe1 in H as [m1 H];
        apply IHe2 in H' as [m2 H'];
        eexists;
        rewrite H, H';
        eauto.
    + apply orb_true_iff in H as [H | H'].
      * apply IHe1 in H as [m1 H]; eauto.
        destruct (possibleValues m e2); rewrite H; eauto.
      * apply IHe2 in H' as [m2 H']; eauto.
        destruct (possibleValues m e1); rewrite H'; eauto.
    + apply noContradictions_equate with δ; dependent destruction e; simpl in *; eauto.
  - eexists; destruct t1; eauto.
Qed.

Lemma unifyVarVar_correct:
  forall τ Δ (δ : env Δ) (m : value_map Δ) (v1 v2: member τ Δ),
  value_map_correct δ m -> hget δ v1 = hget δ v2 ->
  exists m' : value_map Δ, unifyVarVar m v1 v2 = m' /\ value_map_correct δ m'.
Proof.
  unfold unifyVarVar; intros; eexists; split; eauto.
  apply put_intersect_var_var; auto.
Qed.

Lemma unifyVarLit_correct:
  forall τ Δ (δ : env Δ) (m : value_map Δ) (v: member τ Δ) l,
  value_map_correct δ m -> hget δ v = expDenote (Lit l) δ ->
  exists m' : value_map Δ, unifyVarLit m v l = m' /\ value_map_correct δ m'.
Proof.
  unfold unifyVarLit; intros; eexists; split; eauto.
  apply put_intersect_var_val; auto.
Qed.


Lemma equate_correct : forall τ Δ (δ : env Δ) (m : value_map Δ) (e1 e2 : exp Δ τ),
    value_map_correct δ m -> expDenote e1 δ = expDenote e2 δ -> exists m', equate m e1 e2 = Some m' /\ value_map_correct δ m'.
Proof.
  intros.
  dependent destruction e1;
  dependent destruction e2; try (eexists; split; [reflexivity | ]; auto).
  - destruct τ; eexists; simpl; split; autounfold; eauto; apply put_intersect_var_var; auto.
  - eapply unifyVarLit_correct in H; eauto. destruct H; intuition; eexists; split; eauto.
    destruct τ; simpl; subst; reflexivity.
  - destruct b; try (destruct a); simpl; eauto; tauto.
  - destruct u; simpl; eauto.
  - eapply unifyVarLit_correct in H; eauto. destruct H; intuition; eexists; split; eauto.
    destruct τ; simpl; subst; reflexivity.
  - simpl in H0; destruct τ; simpl; subst;
      [rewrite eqb_refl |  rewrite resourceEqb_refl | rewrite PeanoNat.Nat.eqb_refl | destruct t0; simpl | rewrite PeanoNat.Nat.eqb_refl ];
      eauto.
  - destruct b; try destruct a; simpl; eauto.
  - destruct u; simpl; eauto.
Qed.

Lemma equate_notIncorrect: forall τ Δ (δ : env Δ) (m : value_map Δ) (e1 e2 : exp Δ τ),
    value_map_correct δ m ->
    equate m e1 e2 = None ->
    (expDenote e1 δ <> expDenote e2 δ).
Proof.
  dependent destruction e1; dependent destruction e2; intros; try (destruct τ; simpl in H0; discriminate);
  try (destruct t3; simpl in H0; discriminate); try (destruct t2; simpl in H0; discriminate).
  - destruct (tyDenote_dec_eq τ t t0); [subst | ]; dependent destruction τ; simpl in H0 |- *; auto;
      [rewrite eqb_refl in H0
      | rewrite resourceEqb_refl in H0
      | rewrite PeanoNat.Nat.eqb_refl in H0
      | destruct t0; simpl in H0
      | rewrite PeanoNat.Nat.eqb_refl in H0]; discriminate.
Qed.

Lemma possibleValues_correct : forall Δ (δ : env Δ) (m : value_map Δ) (e : exp Δ Bool),
    value_map_correct δ m ->
    expDenote e δ = true -> exists m', possibleValues m e = Some m' /\ value_map_correct δ m'.
Proof.
  dependent induction e; intros; simpl in *; subst; eauto.
  - dependent destruction b; simpl in *; eauto.
    + apply andb_true_iff in H0 as [H1 H2].
      apply IHe1 in H1 as [m1 H1]; auto;
        apply IHe2 in H2 as [m2 H2]; auto;
          destruct H1, H2. rewrite H0, H2; eauto with csl.
    + apply orb_true_iff in H0 as [H1 | H1]; eauto.
      apply IHe1 in H1 as [m1 H1]; auto.
      destruct H1. rewrite H0.
      destruct (possibleValues m e2); try rewrite H0; eauto with csl.
      destruct (possibleValues m e1);
        apply IHe2 in H1 as [m2 H2]; auto; destruct H2;
          try rewrite H1; try rewrite H0; try rewrite Hn; eauto with csl.
    +  apply equate_correct; auto.
       destruct τ; simpl in H0; auto;
       [apply eqb_eq | apply resourceEqb_eq | apply PeanoNat.Nat.eqb_eq | rewrite Bool.eqb_true_iff in H0 | apply PeanoNat.Nat.eqb_eq]; auto.
  - dependent destruction u. eauto.
Qed.

Lemma possibleValues_notIncorrect: forall Δ (δ : env Δ) (m : value_map Δ) (e : exp Δ Bool),
    value_map_correct δ m ->
    possibleValues m e = None ->
    expDenote e δ = false.
Proof.
  dependent induction e; intros.
  - simpl in *; discriminate.
  - simpl in H0; destruct t; simpl; [discriminate |  reflexivity].
  - dependent destruction b; simpl in *;
      remember (possibleValues m e1) as pv1; remember (possibleValues m e2) as pv2; dependent destruction pv1; dependent destruction pv2;
        simpl in *; try discriminate;
          try (apply equate_notIncorrect with (δ:=δ) in H0; auto;
               dependent destruction τ; simpl;
               [ apply eqb_neq | apply resourceEqb_neq | apply PeanoNat.Nat.eqb_neq | apply eqb_false_iff | apply PeanoNat.Nat.eqb_neq]; auto).
      + rewrite IHe2 with (e0:=e2); eauto; apply andb_false_r.
      + rewrite IHe1 with (e0:=e1); eauto.
      + rewrite IHe1 with (e0:=e1); eauto.
      + rewrite IHe1 with (e0:=e1); eauto; rewrite IHe2 with (e0:=e2); eauto.
  - dependent destruction u; simpl in *; discriminate.
Qed.

Ltac rewritePossibleValues :=
  match goal with
  | [ E : possibleValues ?M ?V = _,
      H : match possibleValues ?M ?V with _ => _ end  =  _  |- _ ] => rewrite E in H
  | [ E : possibleValues ?M ?V = _ |-
      match possibleValues ?M ?V with _ => _ end  =  _ ] => rewrite E
  | [ E : _ =  possibleValues ?M ?V |- _ ] => symmetry in E
  end.

Lemma equate_monotone_None:
  forall Δ τ0 (e1 e2 : exp Δ τ0) (m1 m2 : value_map Δ),
    (forall (τ : ty) (x : member τ Δ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    (equate m1 e1 e2 = None <-> equate m2 e1 e2 = None).
Proof.
  dependent induction e1; dependent destruction e2; split; intros; try discriminate;
    try (dependent destruction τ; simpl in *; discriminate);
    try (dependent destruction t3; simpl in *; discriminate);
    try (dependent destruction t2; simpl in *; discriminate).
  destruct (tyDenote_dec_eq τ t t0); [subst | ]; dependent destruction τ; simpl in H0 |- *; auto;
    [ rewrite eqb_refl in H0
    | rewrite resourceEqb_refl in H0
    | rewrite PeanoNat.Nat.eqb_refl in H0
    | try destruct t0; simpl in H0
    | rewrite PeanoNat.Nat.eqb_refl in H0
    | rewrite <- eqb_neq in n
    | rewrite <- resourceEqb_neq in n
    | rewrite <- PeanoNat.Nat.eqb_neq in n
    | apply eqb_false_iff in n
    | rewrite <- PeanoNat.Nat.eqb_neq in n]; try discriminate; rewrite n; reflexivity.
  destruct (tyDenote_dec_eq τ t t0); [subst | ]; dependent destruction τ; simpl in H0 |- *; auto;
    [ rewrite eqb_refl in H0
    | rewrite resourceEqb_refl in H0
    | rewrite PeanoNat.Nat.eqb_refl in H0
    | try destruct t0; simpl in H0
    | rewrite PeanoNat.Nat.eqb_refl in H0
    | rewrite <- eqb_neq in n
    | rewrite <- resourceEqb_neq in n
    | rewrite <- PeanoNat.Nat.eqb_neq in n
    | apply eqb_false_iff in n
    | rewrite <- PeanoNat.Nat.eqb_neq in n]; try discriminate; rewrite n; reflexivity.
Qed.

Lemma equate_monotone_Some:
  forall Δ τ0 (e1 e2 : exp Δ τ0) (m1 m2 m1' m2' : value_map Δ),
    (forall τ (x : member τ _ ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    equate m1 e1 e2 = Some m1' ->
    equate m2 e1 e2 = Some m2' ->
    forall τ (x : member τ _ ), abstract_set_Incl (hget m1' x) (hget m2' x).
Proof.
  dependent induction e1; dependent destruction e2; intros.
  - dependent destruction τ; inversion H0; inversion H1; subst; unfold unifyVarVar; intros;
      apply put_intersect_Incl_var_var; eauto.
  - dependent destruction τ; inversion H0; inversion H1; subst; unfold unifyVarLit; 
      eapply put_intersect_Incl_var_val in H; unfold value_map_Incl; auto.
  - dependent destruction t3; simpl in *; inversion H0; inversion H1; subst; auto.
  - dependent destruction t2; simpl in *; inversion H0; inversion H1; subst; auto.
  - dependent destruction τ; inversion H0; inversion H1; subst; unfold unifyVarLit;
      eapply put_intersect_Incl_var_val in H; unfold value_map_Incl; auto.
  - dependent destruction τ; simpl in *.
      + destruct (t =? t0); inversion H0; inversion H1; subst; auto.
      + destruct (resourceEqb t t0); inversion H0; inversion H1; subst; auto.
      + destruct (Nat.eqb t t0); inversion H0; inversion H1; subst; auto.
      + destruct (Bool.eqb t t0); inversion H0; inversion H1; subst; auto.
      + destruct (Nat.eqb t t0); inversion H0; inversion H1; subst; auto.
  - dependent destruction t3; simpl in *; inversion H0; inversion H1; subst; auto.
  - dependent destruction t2; simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
  - simpl in *; inversion H0; inversion H1; subst; auto.
Qed.

Ltac destructPossibleValues :=
  match  goal with
  | [ E : (match possibleValues ?M ?V with _ => _ end)  =  _  |- _ ] =>
    let v := fresh "eq" in remember (possibleValues M V) as v; dependent destruction v
  | [ |- (match possibleValues ?M ?V with _ => _ end)  =  _ ] =>
    let v := fresh "eq" in remember (possibleValues M V) as v; dependent destruction v                                    
  end; try destructPossibleValues.
  
Lemma possibleValues_monotone_None:
  forall Δ (e : exp Δ Bool) (m1 m2 : value_map Δ),
    (forall τ (x : member τ _ ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    (possibleValues m1 e = None <-> possibleValues m2 e = None).
Proof.
  dependent induction e; split; simpl; intros; try discriminate.
  - destruct t; auto; discriminate.
  - destruct t; auto; discriminate.
  - dependent destruction b; try discriminate;
    repeat destructPossibleValues; simpl in *; try discriminate; repeat rewritePossibleValues; auto.
    * rewrite IHe2 in Heqeq0; eauto; rewrite Heqeq2 in Heqeq0; discriminate.
    * rewrite IHe1 in Heqeq; eauto; rewrite Heqeq in Heqeq0; discriminate.
    * rewrite IHe1 in Heqeq; eauto. rewrite Heqeq in Heqeq1; discriminate.
    * rewrite IHe1 in Heqeq; eauto. rewrite Heqeq in Heqeq1; discriminate.
    * rewrite IHe2 in Heqeq0; eauto. rewrite Heqeq2 in Heqeq0; discriminate.
    * rewrite equate_monotone_None in H0; eauto. 
  - dependent destruction b; try discriminate;
    repeat destructPossibleValues; simpl in *; try discriminate; repeat rewritePossibleValues; auto.
    * rewrite <- IHe2 in Heqeq0; eauto. rewrite Heqeq2 in Heqeq0; discriminate.
    * rewrite <- IHe1 in Heqeq; eauto. rewrite Heqeq in Heqeq0; discriminate.
    * rewrite <- IHe1 in Heqeq; eauto. rewrite Heqeq in Heqeq1; discriminate.
    * rewrite <- IHe1 in Heqeq; eauto. rewrite Heqeq in Heqeq1; discriminate.
    * rewrite <- IHe2 in Heqeq0; eauto. rewrite Heqeq2 in Heqeq0; discriminate.
    * rewrite equate_monotone_None; eauto. 
  - destruct t1; discriminate.
  - destruct t1; discriminate.
Qed.

Lemma possibleValues_monotone_Some :
  forall Δ (e : exp Δ Bool) (m1 m2 m1' m2' : value_map Δ),
    (forall τ (x : member τ _ ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    possibleValues m1 e = Some m1' ->
    possibleValues m2 e = Some m2' ->
    forall τ (x : member τ _ ), abstract_set_Incl (hget m1' x) (hget m2' x).
Proof.
  dependent induction e; simpl; intros.
  - inversion H0; inversion H1; subst; auto.
  - destruct t; inversion H0; inversion H1; subst; auto.
  - dependent destruction b.
    + destructPossibleValues; simpl in *; try discriminate; repeat rewritePossibleValues; auto;
        inversion H0; inversion H1; subst.
      * apply abstract_set_Incl_intersect_all; intros; [eapply IHe1 | eapply IHe2]; eauto.
    + destructPossibleValues; simpl in *; try discriminate; repeat rewritePossibleValues; auto;
        inversion H0; inversion H1; subst.
      * apply abstract_set_Incl_union_all; intros; [eapply IHe1 | eapply IHe2]; eauto.
      * eapply abstract_set_Incl_union_all_l; eauto.
      * rewrite possibleValues_monotone_None with (m2:=m2) in Heqeq1; auto;
          rewrite Heqeq1 in Heqeq; discriminate.
      * rewrite <- possibleValues_monotone_None with (m1:=m1) in Heqeq0; auto.
        rewrite Heqeq2 in Heqeq0; discriminate.
      * eapply IHe1; eauto.
      * rewrite <- possibleValues_monotone_None with (m2:=m2) in Heqeq0; auto.
        rewrite Heqeq2 in Heqeq0; discriminate.
      * rewrite <- possibleValues_monotone_None with (m1:=m1) in Heqeq; auto.
        rewrite Heqeq1 in Heqeq; discriminate.
      * apply possibleValues_monotone_None with (m2:=m2) in Heqeq2; auto.
        rewrite Heqeq0 in Heqeq2; discriminate.
      * eapply IHe2; eauto. 
    + inversion H0; inversion H1; subst; eauto.
    + eapply equate_monotone_Some; eauto.
    + inversion H0; inversion H1; subst; eauto.
  -  dependent destruction u; inversion H0; inversion H1; subst; eauto.
Qed.


Eval compute in possibleValues
                  (full_map [Agent; Agent; Resource; Timestamp])
                  (Binop And
                         (Binop (Eq _ ) (@Lit _ Resource [("a",100)] ) (Var (HNext (HNext (HFirst)))))
                         (Binop (Eq _ ) (@Lit _ Agent "Hans") (Var HFirst))).

Eval compute in possibleValues
                        (full_map [Agent])
                        (@Lit _ Bool false).

Eval compute in possibleValues
                  (full_map [Agent; Agent; Resource; Timestamp])
                  (Binop And
                         (Binop (Eq _ ) (@Lit _ Agent "Hans") (Var HFirst))
                         (Unop Not (Binop (Eq _) (@Lit _ Agent "Hans") (Var HFirst)))).

Definition arg_eval_possible {Δ τ} (e : exp Δ τ) (m : value_map Δ) :=
  match e with
  | Var v => hget m v
  | @Lit _ t l => ActualSet (singleton (tyDenote_dec_eq t) l)
  | _ => FullSet
  end.

Lemma arg_eval_possible_correct{Δ} :
  forall (δ : env Δ) (m : hlist abstract_set Δ) (τ : ty) (e : exp Δ τ),
    value_map_correct δ m ->
    abstract_set_In (expDenote e δ) (arg_eval_possible e m).
Proof.
  intros; dependent destruction e; simpl; intros; intuition.
Qed.

Lemma arg_eval_possible_monotone{Δ} :
  forall (m m': hlist abstract_set Δ) (τ : ty) (e : exp Δ τ),
    (forall τ0 (x : member τ0 _ ), abstract_set_Incl (hget m x) (hget m' x)) ->
    abstract_set_Incl (arg_eval_possible e m) (arg_eval_possible e m').
Proof.
  intros; dependent destruction e; simpl; auto.
Qed.

Lemma possibleValues_None: forall Δ (δ : env Δ) (m : value_map Δ) (e : exp Δ Bool),
    possibleValues m e = None /\  value_map_correct δ m ->
    expDenote e δ = false.
Proof.
  intuition; apply possibleValues_notIncorrect with (m:=m); eauto.
Qed.

Lemma possibleValues_Some: forall Δ (δ : env Δ) (m m' : value_map Δ) (e : exp Δ Bool),
    value_map_correct δ m /\ possibleValues m e = Some m' /\  expDenote e δ = true ->
    value_map_correct δ m'.
Proof.
  intuition. apply possibleValues_correct with (e:=e) in H0; eauto.
  destruct H0; intuition. rewrite H1 in H; inversion H; subst; assumption.
Qed.

Lemma  possibleValues_monotone:
  forall Δ (m1 m2 : value_map Δ) (m1' m2' : value_map (new_var Δ)) (e : exp (new_var Δ) Bool),
    (forall τ (x : member τ _ ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    possibleValues (new_var_value_map m1) e = Some m1' ->
    possibleValues (new_var_value_map m2) e = Some m2' ->
    forall τ (x : member τ (new_var Δ)), abstract_set_Incl (hget m1' x) (hget m2' x).
Proof.
  intros; unfold new_var in *.
  apply possibleValues_monotone_Some with
      (m1 := (HCons FullSet (HCons FullSet (HCons FullSet (HCons FullSet m1)))))
      (m2 := (HCons FullSet (HCons FullSet (HCons FullSet (HCons FullSet m2))))) (e:=e) (τ:=τ) (x:=x); eauto.
  do 4 (eapply abstract_set_Incl_HCons; eauto with csl; intros).
Qed.

Lemma possibleValues_monotone':
  forall Δ (e : exp (new_var Δ) Bool) (m1 m2 : value_map Δ),
    (forall τ (x : member τ _ ), abstract_set_Incl (hget m1 x) (hget m2 x)) ->
    possibleValues (new_var_value_map m1) e = None ->
    possibleValues (new_var_value_map m2) e = None.
Proof.
  intros.
  rewrite possibleValues_monotone_None in H0; eauto.
  intros. unfold new_var_value_map.
  do 4 (eapply abstract_set_Incl_HCons; eauto with csl; intros).
Qed.

Program Instance possible_values_predicate_analysis : @PredicateAnalysis _ _ abstract_set_setlattice :=
  {
    arg_eval := @arg_eval_possible;
    arg_eval_correct := @arg_eval_possible_correct;
    arg_eval_monotone := @arg_eval_possible_monotone;
    analysis _ p m := possibleValues p m;
    analysis_None _ p m δ := possibleValues_None p;
    analysis_Some _ p m m' δ := possibleValues_Some p;
    analysis_monotone := @possibleValues_monotone;
    analysis_monotone' _ m1 m2 p := possibleValues_monotone' p m1 m2
  }.
Next Obligation.
  destruct (possibleValues m p); [left | right]; eauto.
Defined.
Next Obligation.
  unfold new_var_value_map. rewrite x0. symmetry.
  apply <- possibleValues_monotone_None; eauto.
  intros.
  do 4 (eapply abstract_set_Incl_HCons; eauto with csl; intros).
Defined.
Next Obligation.
  unfold new_var_value_map. rewrite x0. symmetry.
  apply <- possibleValues_monotone_None; eauto.
  intros.
  do 4 (eapply abstract_set_Incl_HCons; eauto with csl; intros).
Defined.
