Require Import List Program String ZArith ListSet.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.AbstractValues
        CSL.Analysis.SingletonSet
        CSL.Analysis.Interval.

Set Implicit Arguments.

Definition fairness_result := tyDenote Agent -> interval.

Definition bot : fairness_result := fun _ => EmptyInterval.
Definition success_fr : fairness_result := fun _ => singleton_interval 0.

Inductive direction := Sender | Receiver.

(* Dummy valuation where everything is worth 1 unit *)
Definition valuation (v : tyDenote Resource) (d : direction) : Z :=
  match d with
  | Sender => (0 - Z.of_nat (List.fold_right (fun kv =>  fun s => snd kv + s) 0 v))
  | Receiver => Z.of_nat (List.fold_right (fun kv =>  fun s => snd kv + s) 0 v)
  end.

Definition ev_to_fr (e : event) : fairness_result :=
  match e with
  | Event a1 a2 r _ => fun a => if String.eqb a1 a2 then singleton_interval 0 else
                              if String.eqb a a1 then singleton_interval (valuation r Sender) else
                                if String.eqb a a2 then singleton_interval (valuation r Receiver) else
                                  singleton_interval 0
  end.

Definition add_fr (f1 f2 : fairness_result) : fairness_result := fun a => plus_interval (f1 a) (f2 a).
Definition join_fr (f1 f2 : fairness_result) : fairness_result := fun a => join_interval (f1 a) (f2 a).

Definition Incl_fr (f1 f2 : fairness_result) : Prop :=
  forall a, Incl_interval (f1 a) (f2 a).

Definition β_fairness_folder (e : event) (r : fairness_result) : fairness_result :=
  add_fr (ev_to_fr e) r.

Hint Unfold success_fr add_fr join_fr Incl_fr bot β_fairness_folder : csl.

Definition β_fairness (t : trace) : fairness_result :=
  fold_right β_fairness_folder success_fr t.

Hint Resolve eq_interval_sym plus_interval_l_zero plus_interval_r_zero plus_interval_cong_assoc plus_interval_cong_assoc_comm : csl.

Lemma β_fairness_app_add_eq : forall (t1 t2 t3 : trace) (a : tyDenote Agent),
    appends t1 t2 t3 ->
    forall a, eq_interval ((β_fairness t3) a) ((add_fr (β_fairness t1) (β_fairness t2)) a).
Proof.
  induction 2; simpl; intros; autounfold with csl; try ring.
    destruct h; autounfold; intros;
    destruct (tyDenote_dec_eq Agent t0 t1); subst; eauto with csl.
Qed.

Lemma β_fairness_interleaves_add_eq : forall (t1 t2 t3 : trace) (a : tyDenote Agent),
    interleave t1 t2 t3 ->
    forall a, eq_interval ((β_fairness t3) a) ((add_fr (β_fairness t1) (β_fairness t2)) a).
Proof.
  induction 2; simpl; autounfold with csl; intros; [ | | destruct e | destruct e ];
    intros; [ring | ring | | ];
      destruct (tyDenote_dec_eq Agent t t0); subst; eauto with csl.
Qed.

Definition value_set (sr : abstract_set Resource) (d : direction) : interval :=
  match sr with
  | FullSet => FullInterval
  | ActualSet s => set_fold_right
                    (fun r i => join_interval (singleton_interval (valuation r d)) i)
                    s EmptyInterval
  end.

Ltac value_set_solver :=
  intuition;
  match goal with
  | |- Incl_interval EmptyInterval _ => apply Incl_interval_empty
  | |- Incl_interval (join_interval ?I ?I) ?I => apply Incl_interval_eq; apply join_interval_idem
  | |- In_interval _ (value_set (ActualSet (_ :: _)) _)  => apply In_interval_join_intro
  | |- Incl_interval (value_set (ActualSet (_ :: _)) _) _  =>
    etransitivity; [apply Incl_interval_monotone | ]
  | [H : In_interval (valuation _ _ ) (value_set _ _) |- _ ] =>
    let h:=fresh "H" in
    apply In_interval_Incl_interval in H as h
  end; simpl; intuition; eauto with csl.

Lemma value_set_correct : forall r d sr,
    abstract_set_In r sr ->
    In_interval (valuation r d) (value_set sr d).
Proof.
  intros;destruct sr; auto;
    induction s; [inversion H | destruct H; subst ]; value_set_solver.
Qed.

Lemma valuation_set_In : forall x s d,
    set_In x s ->
    Incl_interval (singleton_interval (valuation x d)) (value_set (ActualSet s) d).
Proof.
  intros; induction s; [inversion H | ];
    destruct H; subst; autounfold with csl; value_set_solver.
Qed.

Lemma abstract_set_Incl_cons : forall τ a (s s0 : set (tyDenote τ)),
    abstract_set_Incl (ActualSet (a :: s)) (ActualSet s0) ->
    abstract_set_Incl (ActualSet s) (ActualSet s0) /\ set_In a s0.
Proof.
  simpl; intuition.
Qed.

Lemma value_set_monotone : forall sr sr' d,
    abstract_set_Incl sr sr' ->
    Incl_interval (value_set sr d) (value_set sr' d).
Proof.
  intros; destruct sr, sr'; auto with csl; [contradiction | ]; induction s; [value_set_solver | ];
  apply abstract_set_Incl_cons in H; intuition;
  apply valuation_set_In with (d := d) in H1;
  transitivity (join_interval (value_set (ActualSet s0) d) (value_set (ActualSet s0) d));
  value_set_solver.
Qed.

Definition value_transfer_approx (sr : abstract_set Resource) (sa : set (tyDenote Agent)) (d : direction) : fairness_result :=
  set_fold_right (fun a fr a' => if String.eqb a a'
                              then join_interval (singleton_interval 0) (value_set sr d)
                              else fr a')
                 sa success_fr.

Ltac destruct_eqb a b :=
  let Hneq := fresh "Hneq" in
  destruct (tyDenote_dec_eq Agent a b) as [? | Hneq]; subst;
  [try rewrite eqb_refl in * | try rewrite <- eqb_neq in Hneq; try rewrite Hneq in *; rewrite eqb_sym in Hneq; try rewrite Hneq in * ].

Hint Resolve In_interval_join_intro Incl_interval_full : csl.
Hint Resolve fact1 fact2 fact3 fact4 fact1' fact2' fact3' fact4' fact5
     In_interval_Incl_interval In_interval_zero_plus In_interval_plus_zero
     In_interval_zero_plus' In_interval_zero_plus'' In_interval_join_intro
     In_interval_join_intro1 In_interval_join_intro2 In_interval_plus
     fact6 fact7 Incl_interval_eq : csl.

Lemma value_transfer_approx_notIn : forall a sr sa d,
    ~ set_In a sa ->
    eq_interval ((value_transfer_approx sr sa d) a) (singleton_interval 0).
Proof.
  induction sa; intros; intuition.
  - simpl; auto.
  - etransitivity; simpl; [destruct_eqb a0 a; simpl in *; [intuition | rewrite IHsa] | ]; eauto with csl.
Qed.

Lemma value_transfer_approx_correct : forall r a sr sa d,
    set_In a sa ->
    In_interval (valuation r d) (value_set sr d) ->
    Incl_interval (singleton_interval (valuation r d)) ((value_transfer_approx sr sa d) a) /\
    Incl_interval (singleton_interval 0) ((value_transfer_approx sr sa d) a).
Proof.
  intros;
  induction sa; [ inversion H | destruct H; subst ].
  - unfold value_transfer_approx, set_fold_right, fold_right;
    rewrite eqb_refl; split; interval_solver idtac.
  - Opaque join_interval.
    apply IHsa in H as [? ?]; simpl; destruct_eqb a0 a; split; auto; interval_solver idtac.
Qed.

Definition value_transfer (sr : abstract_set Resource) (sa : abstract_set Agent) (d : direction) : fairness_result :=
  match sa with
  | FullSet => fun _ => join_interval (singleton_interval 0) (value_set sr d)
  | ActualSet [] => fun _ => EmptyInterval
  | ActualSet (a :: xs) => fun a' => if (is_singleton_set (a :: xs)) then (if String.eqb a a' then value_set sr d
                                                                           else singleton_interval 0)
                                     else value_transfer_approx sr (a :: xs) d a'
  end.

Ltac eq_singleton a :=
  match goal with
    [H : true = is_singleton_set ( ?T :: _)  |- _ ] =>
    assert (T = a) by
        (simpl in H |- ;
         symmetry in H;
         intuition;
         apply is_singleton_set_In_eq with (v:=a) in H; simpl; intuition);
    subst; simpl in H |- *; try rewrite eqb_refl; try rewrite <- H; auto with csl
  | [ H : false = is_singleton_set (?T :: _ ) |- _ ] =>
    simpl in H; simpl; rewrite <-H;
    destruct_eqb T a ; auto with csl
  end.

(* This correctness statement is not strong enough *)
Lemma value_transfer_correct : forall r a sr sa d,
    abstract_set_In r sr ->
    abstract_set_In a sa ->
    Incl_interval (singleton_interval (valuation r d)) ((value_transfer sr sa d) a).
Proof.
  intros.
  apply value_set_correct with (d := d) in H.
  value_set_solver; destruct sa.
  - value_set_solver.
  - destruct s; [contradiction | ];
    remember (is_singleton_set (t::s)) as b; destruct b; eq_singleton a;
    apply value_transfer_approx_correct; simpl in H0; intuition;
      subst; rewrite eqb_refl in Hneq; discriminate.
Qed.

Definition transfer_analysis Δ (m : value_map (new_var Δ)) : fairness_result :=
  let a1 := hget m HFirst in
  let a2 := hget m (HNext HFirst) in
  let r := hget m (HNext (HNext HFirst)) in
  add_fr (value_transfer r a1 Sender) (value_transfer r a2 Receiver).

Definition abstract_set_In_dec : forall τ (a : tyDenote τ) x,
    {abstract_set_In a x} + {~ abstract_set_In a x}.
Proof.
  intros; destruct x; [ left; simpl; auto | simpl];
  apply set_In_dec; apply tyDenote_dec_eq.
Defined.

Lemma not_in_singleton : forall a sa sr r d,
    ~ abstract_set_In a (ActualSet sa) ->
    abstract_set_In r sr ->
    eq_interval (singleton_interval 0) (value_transfer_approx sr sa d a).
Proof.
  intros; apply value_set_correct with (d := d) in H0 as H1;
  induction sa; [simpl; f_equal; auto | ];
  simpl in *;
  apply Decidable.not_or in H as [? ?];
  rewrite <- eqb_neq in H;
  rewrite H;
  apply IHsa in H2; auto.
Qed.

Ltac value_transfer_approx_solver A Dir :=
  match goal with
  | [H : set_In A ?S,
     H' : In_interval (valuation ?R Dir) (value_set ?SR Dir)
     |- _ ] =>
    eapply value_transfer_approx_correct with (r:=R) (sr :=SR) (d := Dir) in H as [? ?]
  | [H : abstract_set_In A ?S,
     H' : In_interval (valuation ?R Dir) (value_set ?SR Dir)
     |- _ ] =>
    eapply value_transfer_approx_correct with (r:=R) (sr :=SR) (d := Dir) in H as [? ?]
  | [H : set_In A ?S,
     H1: In_interval (Z.of_nat (fold_right _ 0 ?R)) (value_set ?SR Dir)
     |- _ ] =>
    apply value_transfer_approx_correct with (r:=R) (sr :=SR) (d := Dir) in H as [? ?]
  | [H : abstract_set_In A ?S,
     H1: In_interval  (Z.of_nat (fold_right _ 0 ?R)) (value_set ?SR Dir)
     |- _ ] =>
    apply value_transfer_approx_correct with (r:=R) (sr :=SR) (d := Dir) in H as [? ?]
  end; auto with csl.
Ltac eautocsl := (eauto with csl).

Lemma transfer_analysis_correct : forall Δ e (δ : env Δ) (m' : value_map (new_var Δ)) a,
    value_map_correct (addEvent e δ) m' ->
    Incl_interval (ev_to_fr e a) (transfer_analysis m' a).
Proof.
  intros;
  unfold transfer_analysis, add_fr in *;
  destruct e as [a1 a2 r ?];
  do 3 (dependent destruction m').

  remember (H Agent HFirst) as H1;
  remember (H Agent (HNext HFirst)) as H2;
  remember (H Resource (HNext (HNext HFirst))) as H3;
  simpl;
  simpl in H, H1, H2, H3;
  clear HeqH1 HeqH2 HeqH3 H m' δ Δ t.

  rename a1 into a__sender;
  rename a2 into a__receiver;
  rename a into sa__sender;
  rename a0 into sa__receiver;
  rename a3 into sr;
  rename a4 into a;
  apply value_set_correct with (d := Sender) in H3 as H4;
  apply value_set_correct with (d := Receiver) in H3 as H5.

  destruct sa__sender, sa__receiver.

  -  (* FullSet, FullSet *)
    destruct_eqb a a__sender;
      destruct_eqb a__sender a__receiver;
      try (destruct_eqb a a__receiver); interval_solver simpl.

  - (* FullSet, ActualSet _ *)
    simpl in H2; destruct s; [inversion H2 | ];
      remember (is_singleton_set (t :: s)); destruct b.
    + (* subcase: the ActualSet (t::s) is a singleton set *)
      eq_singleton a__receiver.
      destruct_eqb a a__sender.
      * destruct_eqb a__sender a__receiver; simpl in H4, H5; ring_simplify; interval_solver eauto.
      * destruct_eqb a a__receiver.
        -- rewrite Hneq; interval_solver idtac.
        -- ring_simplify; destruct_eqb a__sender a__receiver; auto with csl.
    + (* subcase: the ActualSet (t::s) is not a singleton set *)
      simpl in Heqb; simpl; rewrite <- Heqb;
      destruct (abstract_set_In_dec a (ActualSet (t :: s))).
      * destruct_eqb a a__sender.
        -- destruct_eqb a__sender a__receiver;
             [value_transfer_approx_solver a__receiver Receiver |
              value_transfer_approx_solver a__sender Receiver];
             ring_simplify; interval_solver idtac.
        -- value_transfer_approx_solver a Receiver.
           simpl in *.
           destruct_eqb a a__receiver;
             destruct_eqb a__sender a__receiver; ring_simplify; auto with csl.
      * apply not_in_singleton with (d := Receiver) (sr := sr) (r := r) in n as H'; auto.
        rewrite <- H'.
        transitivity (join_interval (singleton_interval 0) (value_set sr Sender)).
         -- destruct_eqb a__sender a__receiver; auto with csl;
              destruct_eqb a__sender a; [interval_solver idtac | ];
              destruct_eqb a__receiver a; intuition.
         -- interval_solver idtac.

  - (* ActualSet _, FullSet *)
    destruct s; [ inversion H1 | ];
      remember (is_singleton_set (t :: s)); destruct b.

    + (* subcase: ActualSet (t::s) is a singleton set *)
      eq_singleton a__sender;
      destruct_eqb a a__sender.
      * destruct_eqb a__sender a__receiver; interval_solver eauto.
      * destruct_eqb a a__receiver.
        -- rewrite Hneq; ring_simplify; interval_solver idtac.
        -- destruct_eqb a__sender a__receiver; ring_simplify; auto with csl.

    + (* subcase: ActualSet (t::s) is not a singleton set *)
      simpl in Heqb; simpl; rewrite <- Heqb;
       destruct (abstract_set_In_dec a (ActualSet (t :: s))).
      * value_transfer_approx_solver a Sender;
          destruct_eqb a a__sender; destruct_eqb a__sender a__receiver;
            [ | | | destruct_eqb a a__receiver]; interval_solver idtac.
      * apply not_in_singleton with (d := Sender) (sr := sr) (r := r) in n as H'; auto;
        destruct_eqb a__sender a__receiver.
        -- rewrite <- H'; ring_simplify; interval_solver idtac.
        -- destruct_eqb a__receiver a.
           ++ rewrite Hneq; rewrite <- H'; ring_simplify; interval_solver idtac.
           ++ destruct_eqb t a; destruct_eqb a a__sender; [elim n | | contradiction | ]; auto with csl;
                simpl in *;
                 apply not_in_singleton with (d := Sender) (sr := sr) (r := r) (a:=a) (sa:=s) in H3; auto;
                   rewrite <- H3; ring_simplify; interval_solver idtac.

  - (* ActualSet _, ActualSet _ *)
    destruct s; [ inversion H1 | ];
    destruct s0; [contradiction | ].
    remember (is_singleton_set (t :: s)).
    remember (is_singleton_set (t0 :: s0)).
    destruct b; destruct b0.

    + (* two singleton sets *)
      eq_singleton a__sender. eq_singleton a__receiver.
      destruct_eqb a a__sender;
        destruct_eqb a__receiver a__sender.
      * interval_solver eauto.
      * ring_simplify; interval_solver simpl.
      * destruct_eqb a__sender a; [discriminate | auto with csl].
      * destruct_eqb a__receiver a; auto with csl; ring_simplify; interval_solver idtac.

    + (* singleton set, approx set *)
      eq_singleton a__sender;  destruct s0; [discriminate | ];
      simpl in Heqb0; simpl; rewrite <- Heqb0;
      destruct (abstract_set_In_dec a (ActualSet (t0 :: t :: s0))).
      * value_transfer_approx_solver a Receiver;
        unfold valuation in *; interval_solver idtac;
          destruct_eqb a a__sender; destruct_eqb a__receiver a__sender;
            [simpl in H4, H5 | | | destruct_eqb a a__receiver; destruct_eqb t0 a__receiver];
            ring_simplify; interval_solver eautocsl.
              destruct_eqb t a__receiver; [auto with csl | ];
                assert (set_In a__receiver s0) by
                  (simpl in *; intuition; subst; rewrite eqb_refl in *; discriminate); 
                value_transfer_approx_solver a__receiver Receiver; interval_solver idtac.
      * assert (H: ~ abstract_set_In a (ActualSet s0)) by (simpl in n; intuition);
        apply not_in_singleton with (d := Receiver) (sr := sr) (r := r) in H as H'; auto.
        assert (t0 <> a) by (simpl in n; intuition);
          assert (t <> a) by (simpl in n; intuition);
          destruct_eqb t0 a; [ contradiction | ];
            destruct_eqb t a; [ contradiction | ];
              destruct_eqb a__receiver a__sender.
        -- rewrite <- H'; ring_simplify; destruct_eqb a__sender a; [intuition | auto with csl].
        -- destruct_eqb a a__receiver; [contradiction | ];
             destruct_eqb a a__sender; rewrite <- H'; ring_simplify; interval_solver eautocsl.

    + (* approx set, singleton set *)
      eq_singleton a__receiver;  destruct s; [discriminate | ];
      simpl in Heqb; simpl; rewrite <-Heqb;
      destruct (abstract_set_In_dec a (ActualSet (t :: t0 :: s))).
      * value_transfer_approx_solver a Sender; unfold valuation in *;
          destruct_eqb a a__receiver; destruct_eqb a__receiver a__sender;
            simpl in *; ring_simplify; [ | | | destruct_eqb a a__sender];
              interval_solver eauto.
      * assert (H1' := H1).
        value_transfer_approx_solver a__sender Sender.
        assert (Hn: ~ abstract_set_In a (ActualSet s)) by (simpl in n; intuition);
        apply not_in_singleton with (d := Sender) (sr := sr) (r := r) in n as H'; auto;
        apply not_in_singleton with (d := Sender) (sr := sr) (r := r) in Hn as H''; auto;
          assert (t0 <> a) by (simpl in n; intuition);
          assert (t <> a) by (simpl in n; intuition);
          destruct_eqb a t0; [ contradiction | ];
            destruct_eqb a t; [ contradiction | ];
              destruct_eqb a__receiver a__sender;
              destruct_eqb a__sender a; [simpl in H, H2; intuition | ring_simplify; auto with csl | contradiction | ];
                destruct_eqb a__receiver a; rewrite <- H''; ring_simplify; interval_solver eautocsl.

    + (* two approx sets *)
      destruct s; [discriminate | destruct s0; [discriminate | ] ];
      simpl in Heqb, Heqb0; simpl; rewrite <-Heqb; rewrite <- Heqb0;
        destruct (abstract_set_In_dec a (ActualSet (t :: t1 :: s)));
        destruct (abstract_set_In_dec a (ActualSet (t0 :: t2 :: s0))).
      *   value_transfer_approx_solver a Receiver;
            value_transfer_approx_solver a Sender;
            destruct_eqb a a__sender; destruct_eqb a__sender a__receiver;
              [ |  value_transfer_approx_solver a__sender Receiver | destruct_eqb a a__receiver; [discriminate | ] | destruct_eqb a a__receiver];
              interval_solver idtac.
      * destruct_eqb a a__receiver; [ intuition  | ];
        apply not_in_singleton with (sr := sr) (r := r) (d := Receiver) in n; auto; rewrite <- n;
          value_transfer_approx_solver a Sender; destruct_eqb a__sender a__receiver;
            [ | destruct_eqb a a__sender]; ring_simplify; interval_solver idtac.
      * value_transfer_approx_solver a Receiver; destruct_eqb a__sender a__receiver;
          [ | destruct_eqb a a__sender; [intuition | ]]; destruct_eqb a a__receiver;
            apply not_in_singleton with (sr := sr) (r := r) (d := Sender) in n; auto; rewrite <- n; ring_simplify;
              [interval_solver intuition | | | ]; interval_solver idtac.
      * destruct_eqb a a__receiver;
          [ destruct_eqb a__sender a__receiver; intuition |  destruct_eqb a a__sender; [ intuition | ]];
        apply not_in_singleton with (sr := sr) (r := r) (d := Sender) in n; auto;
        apply not_in_singleton with (sr := sr) (r := r) (d := Receiver) in n0; auto;
        rewrite <- n; rewrite <- n0; destruct_eqb a__sender a__receiver; ring_simplify; auto with csl.
Qed.

Lemma value_transfer_approx_over : forall a s sr d,
    set_In a s ->
    Incl_interval (value_set sr d) (value_transfer_approx sr s d a)
    /\
    Incl_interval (singleton_interval 0) (value_transfer_approx sr s d a).
Proof.
  intros; induction s; [contradiction | destruct H; subst]; simpl;[rewrite eqb_refl | destruct_eqb a0 a]; auto with csl.
Qed.

Lemma value_transfer_approx_over' : forall a s sr d,
    Incl_interval (value_transfer_approx sr s d a) (join_interval (singleton_interval 0) (value_set sr d)).
Proof.
  intros; induction s; simpl in *; [ | destruct_eqb a0 a]; auto with csl.
Qed.

Lemma value_transfer_approx_zero : forall a s sr d,
    Incl_interval (singleton_interval 0) (value_transfer_approx sr s d a).
Proof.
  intros; induction s; simpl in *; [ | destruct_eqb a0 a]; auto with csl.
Qed.

Lemma value_transfer_approx_Incl_0_value_set: forall sr sr' d sa a,
  Incl_interval (value_set sr d) (value_set sr' d) ->
  Incl_interval (value_transfer_approx sr sa d a) (join_interval (singleton_interval 0) (value_set sr' d)).
Proof.
  intros; transitivity (join_interval (singleton_interval 0) (value_set sr d));
    [ apply value_transfer_approx_over' |  apply Incl_interval_r_monotone; auto].
Qed.

Lemma Incl_interval_eq_interval_r: forall i1 i2 i,
  eq_interval i1 i2 -> Incl_interval i i1 -> Incl_interval i i2.
Proof.
  intros; rewrite H in H0; assumption.
Qed.

Definition is_empty_set{A} (s : set A) :=
  match s with
  | [] => True
  | _ => False
  end.

Lemma value_transfer_notIn: forall s sr d a,
    is_singleton_set s = false ->
    ~ is_empty_set s ->
    ~ abstract_set_In a (ActualSet s) ->
    eq_interval (value_transfer sr (ActualSet s) d a) (singleton_interval 0).
Proof.
  induction s; intros; intuition.
  - simpl in *; contradiction.
  - destruct s; [discriminate | ];
    simpl in H |- *; rewrite H;
    destruct_eqb a a0.
    + simpl in *; intuition.
    + destruct_eqb t a0; [simpl in H1; intuition | ];
      apply value_transfer_approx_notIn; simpl in H1; intuition.
Qed.

Lemma value_transfer_monotone : forall sa sa' sr sr' d a,
    abstract_set_Incl sa sa' ->
    abstract_set_Incl sr sr' ->
    Incl_interval (value_transfer sr sa d a)
                  (value_transfer sr' sa' d a).
Proof.
  intros.
  destruct sa, sa'; try (simpl; progress auto);
    try (simpl; easy);
    apply value_set_monotone with (d := d) in H0.
  - unfold value_transfer;
      apply Incl_interval_monotone; auto with csl.
  - destruct s.
    + apply Incl_interval_empty.
    + remember (is_singleton_set (t::s)); destruct b;
      simpl in Heqb; simpl; rewrite <- Heqb; destruct_eqb t a; auto with csl.
        * rewrite H0; auto with csl.
        * rewrite <- H0; apply value_transfer_approx_over'.
  - destruct s; [apply Incl_interval_empty | ]; remember (is_singleton_set (t::s)); destruct b.
    + assert (forall v, set_In v (t::s) -> v = t) by (apply is_singleton_set_In_eq; auto);
        simpl in Heqb; simpl; rewrite <- Heqb;
        simpl in H; destruct s0; [edestruct H; intuition | ];
        remember (is_singleton_set (t0::s0)); destruct b.
        * assert (forall v, set_In v (t0::s0) -> v = t0) by (apply is_singleton_set_In_eq; auto);
          simpl in Heqb0; rewrite <- Heqb0; simpl in H, H1, H2;
            specialize H with t; specialize H1 with t; specialize H2 with t;
              destruct_eqb t a; destruct_eqb t0 a; auto; intuition;
                subst; rewrite eqb_refl in *; discriminate.
        *  simpl in Heqb0; rewrite <- Heqb0; simpl in H, H1;
            destruct_eqb t a; destruct_eqb t0 a; auto with csl; [ | apply value_transfer_approx_zero];
              rewrite H0; apply value_transfer_approx_over;
                specialize H with a; intuition; subst; rewrite eqb_refl in *; discriminate.
      + assert (exists v, set_In v (t :: s) /\ t <> v) by (eapply not_singleton; auto);
        simpl in Heqb; simpl; rewrite <- Heqb; destruct s0.
        * simpl in H; edestruct H; eauto.
        * remember (is_singleton_set (t0::s0)). destruct b.
           -- (* Impossible: a non-singleton set is a subset of a singleton set *)
              assert (forall v, set_In v (t0::s0) -> v = t0) by (apply is_singleton_set_In_eq; auto).
              do 2 destruct H1; simpl in H, H1, H2.
              assert (t0 = x \/ set_In x s0) by auto.
              assert (t0 = t \/ set_In t s0) by auto.
              destruct H1; [contradiction | ].
              destruct H4; [subst; destruct H5; [subst; contradiction | specialize H  with t; intuition] | ];
              assert (forall v : string, t0 = v \/ set_In v s0 -> v = t0) by auto;
              specialize H2 with x; intuition; subst; intuition.
           -- (* Neither of the sets is a singleton *)
             destruct (abstract_set_In_dec a (ActualSet (t0::s0))).
             ++ simpl in Heqb0; rewrite <- Heqb0;
                destruct_eqb t0 a; destruct_eqb t a.
                --- apply Incl_interval_r_monotone; assumption.
                --- etransitivity;
                          [apply value_transfer_approx_over'| apply Incl_interval_r_monotone]; assumption.
                --- assert (H2: eq_interval (value_transfer_approx sr' s0 d a)
                                        (join_interval (value_transfer_approx sr' s0 d a)
                                                       (value_transfer_approx sr' s0 d a) ) ) by
                      (symmetry; apply join_interval_idem); rewrite H2;
                    apply Incl_interval_monotone;
                    split; [apply value_transfer_approx_zero | rewrite H0; eapply value_transfer_approx_over];
                    simpl in H; specialize H with a; intuition;
                    subst; rewrite eqb_refl in *; discriminate.
                --- assert (H2: eq_interval (value_transfer_approx sr' s0 d a)
                                        (join_interval (value_transfer_approx sr' s0 d a)
                                                       (value_transfer_approx sr' s0 d a) ) ) by
                      (symmetry; apply join_interval_idem); rewrite H2.
                    transitivity (join_interval (singleton_interval 0) (value_set sr' d));
                          [apply value_transfer_approx_Incl_0_value_set; auto | apply Incl_interval_monotone];
                          split; [apply value_transfer_approx_zero | eapply value_transfer_approx_over];
                            simpl in a0; intuition; subst; rewrite eqb_refl in *; discriminate.
             ++ assert (~abstract_set_In a (ActualSet (t::s))) by (simpl in H, n; simpl; intro; elim n; auto);
                destruct_eqb t a; [simpl in H2; intuition | ];
                assert (is_singleton_set (t0::s0) = false) by (symmetry; auto).
                assert (is_singleton_set (t::s) = false) by (symmetry; simpl; auto).
                apply value_transfer_notIn with (sr:=sr') (d:=d) in n; auto; rewrite n;
                apply Incl_interval_eq; apply value_transfer_approx_notIn; simpl in *; intuition.
Qed.
