Require Import ZArith Psatz Nsatz.
Require Import Relation_Operators
               Operators_Properties.
Require Import Setoid Morphisms.
Require Import CSL.Analysis.Lattice.

Set Implicit Arguments.

Inductive interval : Set :=
| EmptyInterval
| Interval : { ' (x, y) | Z.le x y } -> interval
| Below : Z -> interval
| Above : Z -> interval
| FullInterval : interval.

Ltac interval_killer :=
  repeat match goal with
         | [ s : {'(_, _) : Z * Z | (_ <= _)%Z } |- _ ] => destruct s as [x ?]; destruct x as [? ?]
         end; simpl in *; f_equal; auto; try lia.

Ltac kill_interval i := destruct i; interval_killer.
Ltac kill_2intervals i i' := destruct i, i'; interval_killer.

Definition eq_interval : relation interval :=
  fun a => fun b => match a, b with
  | Interval i1, Interval i2 => proj1_sig i1 = proj1_sig i2
  | _, _ => a = b
  end.

Lemma eq_interval_refl : forall (i : interval),
    eq_interval i i.
Proof.
  destruct i; unfold eq_interval; auto.
Qed.

Lemma eq_interval_sym : forall (a b : interval),
    eq_interval a b -> eq_interval b a.
Proof.
  unfold eq_interval;
  destruct a, b;
  intuition.
Qed.

Lemma eq_interval_trans : forall (a b c : interval),
    eq_interval a b -> eq_interval b c -> eq_interval a c.
Proof.
  intros;
  destruct a, b, c; inversion H; inversion H0;
    unfold eq_interval in *; intuition;
      rewrite H; assumption.
Qed.

Add Relation interval eq_interval
    reflexivity proved by eq_interval_refl
    symmetry proved by eq_interval_sym
    transitivity proved by eq_interval_trans
      as eq_interval_rel.

Definition In_interval (z : Z) (i : interval) : Prop :=
  match i with
  | EmptyInterval => False
  | Interval i' => let (a, b) := proj1_sig i' in (a <= z)%Z /\ (z <= b)%Z
  | Below b => Z.le z b
  | Above a => Z.le a z
  | FullInterval => True
  end.

Lemma In_interval_eq : forall (i1 i2 : interval),
    eq_interval i1 i2 ->
    forall z, In_interval z i1 <-> In_interval z i2.
Proof.
  unfold eq_interval, In_interval; intros;
    destruct i1; destruct i2; inversion H; intuition.
Qed.

Add Morphism In_interval
    with signature Z.eq ==> eq_interval ==> iff
      as In_interval_compat.
Proof.
intros; rewrite H; apply In_interval_eq; assumption.
Qed.

Definition Incl_interval (i1 i2 : interval) : Prop :=
  forall z, In_interval z i1 -> In_interval z i2.

Hint Unfold Incl_interval : csl.

Add Morphism In_interval
    with signature Z.eq ==> Incl_interval ==> Basics.impl
      as In_interval_incl.
Proof.
  autounfold; unfold Basics.impl; intros; rewrite <- H; auto.
Qed.

Lemma Incl_interval_eq : forall (i1 i2 : interval),
    eq_interval i1 i2 ->
    Incl_interval i1 i2.
Proof.
  autounfold with csl;
  intros;
  rewrite <- In_interval_eq; eauto.
Qed.

Lemma Incl_interval_refl : forall i,
    Incl_interval i i.
Proof.
  autounfold with csl; auto.
Qed.

Lemma Incl_interval_trans : forall i1 i2 i3,
    Incl_interval i1 i2 ->
    Incl_interval i2 i3 ->
    Incl_interval i1 i3.
Proof.
  autounfold with csl; intros; auto.
Qed.

Add Relation interval Incl_interval
    reflexivity proved by Incl_interval_refl
    transitivity proved by Incl_interval_trans
      as Incl_interval_rel.

Add Morphism Incl_interval
    with signature eq_interval ==> eq_interval ==> iff
      as Incl_interval_eqmor.
Proof.
  split; intros.
  - transitivity x.
    + symmetry in H; apply Incl_interval_eq in H; assumption.
    + transitivity x0;  [ | apply Incl_interval_eq in H0]; assumption.
  - transitivity y.
    + apply Incl_interval_eq in H; assumption.
    + transitivity y0;
      [ | symmetry in H0; apply Incl_interval_eq in H0]; assumption.
Qed.

Definition singleton_interval (x : Z) : interval :=
  Interval (exist _ (x, x) (Z.le_refl x)).

Lemma In_interval_Incl_interval : forall z i,
    In_interval z i <-> Incl_interval (singleton_interval z) i.
Proof.
  autounfold with csl; intros; intuition; kill_interval i; apply (H z); lia.
Qed.

Definition join_interval (i1 i2 : interval) : interval.
  refine (match i1, i2 with
          | EmptyInterval, i => i
          | i, EmptyInterval => i
          | Interval i1', Interval i2' => Interval _
          | Interval i1', Below d => let (a, b) := proj1_sig i1' in
                                        Below (Z.max b d)
          | Interval i1', Above c => let (a, b) := proj1_sig i1' in
                                        Above (Z.min a c)
          | Below b, Interval i2' => let (c, d) := proj1_sig i2' in
                                        Below (Z.max b d)
          | Above a, Interval i2' => let (c, d) := proj1_sig i2' in
                                        Above (Z.min a c)
          | Below b, Below d => Below (Z.max b d)
          | Above a, Above c => Above (Z.min a c)
          | Below _, Above _ => FullInterval
          | Above _, Below _ => FullInterval
          | FullInterval, _ => FullInterval
          | _, FullInterval => FullInterval
          end).
  destruct i1', i2'.
  destruct x, x0.
  exists (Z.min z z1, Z.max z0 z2).
  destruct (Z.min_dec z z1);
    destruct (Z.max_dec z0 z2);
    rewrite e, e0.
  exact y.
  apply Z.min_l_iff in e.
  apply Z.max_r_iff in e0.
  apply Z.le_trans with (m := z1).
  exact e. exact y0.
  apply Z.min_r_iff in e.
  apply Z.max_l_iff in e0.
  apply Z.le_trans with (m := z2).
  exact y0. exact e0.
  exact y0.
Defined.

Lemma join_interval_l_id : forall i,
    eq_interval (join_interval EmptyInterval i) i.
Proof.
  reflexivity.
Qed.

Lemma join_interval_r_id : forall i,
    eq_interval (join_interval i EmptyInterval) i.
Proof.
  kill_interval i.
Qed.

Lemma join_interval_l_full : forall i,
    eq_interval (join_interval FullInterval i) FullInterval.
Proof.
  kill_interval i.
Qed.

Lemma join_interval_r_full : forall i,
    eq_interval (join_interval i FullInterval) FullInterval.
Proof.
  kill_interval i.
Qed.

Lemma join_interval_comm : forall i1 i2,
    eq_interval (join_interval i1 i2) (join_interval i2 i1).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma join_interval_assoc : forall i1 i2 i3,
    eq_interval (join_interval (join_interval i1 i2) i3) (join_interval i1 (join_interval i2 i3)).
Proof.
  kill_interval i1; kill_2intervals i2 i3.
Qed.

Lemma join_interval_idem : forall i,
    eq_interval (join_interval i i) i.
  kill_interval i.
Qed.

Add Morphism join_interval
    with signature eq_interval ==> eq_interval ==> eq_interval
      as join_interval_eqmor.
Proof.
  intros.
  destruct x; destruct x0; simpl in *; subst; auto.
  - destruct y; simpl; inversion H; auto.
  - destruct y; destruct y0; try discriminate;
    destruct s as [[x0 x0'] x1], s1 as [[y0 y0'] y1]; simpl in *; subst;
    destruct s0 as [[x2 x2'] x3 ], s2 as [[y2 y2'] y3 ]; simpl in *;
    inversion H0; inversion H; subst; reflexivity.
  - destruct y; simpl; inversion H; auto; reflexivity.
  - destruct y; simpl; inversion H; auto; reflexivity.
  - destruct y; simpl; inversion H; auto; reflexivity.
  - destruct y0; simpl; inversion H0; auto; reflexivity.
  - destruct y0; simpl; inversion H0; auto; reflexivity.
  - destruct y0; simpl; inversion H0; auto; reflexivity.
Qed.

Definition plus_interval (i1 i2 : interval) : interval.
  refine (match i1, i2 with
          | EmptyInterval, _ => EmptyInterval
          | _, EmptyInterval => EmptyInterval
          | FullInterval, _ => FullInterval
          | _, FullInterval => FullInterval
          | Interval i1', Interval i2' => Interval _
          | Interval i1', Below d => let (a, b) := proj1_sig i1' in
                                        Below (b + d)
          | Interval i1', Above c => let (a, b) := proj1_sig i1' in
                                        Above (a + c)
          | Below b, Interval i2' => let (c, d) := proj1_sig i2' in
                                        Below (b + d)
          | Above a, Interval i2' => let (c, d) := proj1_sig i2' in
                                        Above (a + c)
          | Below b, Below d => Below (b + d)
          | Above a, Above c => Above (a + c)
          | Below _, Above _ => FullInterval
          | Above _, Below _ => FullInterval
          end).
  destruct i1', i2'.
  destruct x, x0.
  exists ((z + z1)%Z, (z0 + z2)%Z).
  apply Z.add_le_mono.
  exact y. exact y0.
Defined.

Lemma plus_interval_r_zero : forall i, eq_interval (plus_interval i (singleton_interval 0)) i.
Proof.
  kill_interval i.
Qed.

Lemma plus_interval_l_zero : forall i, eq_interval (plus_interval (singleton_interval 0) i) i.
Proof.
  kill_interval i.
Qed.

Lemma plus_interval_comm : forall i1 i2,
    eq_interval (plus_interval i1 i2) (plus_interval i2 i1).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma plus_interval_assoc : forall i1 i2 i3,
    eq_interval (plus_interval (plus_interval i1 i2) i3) (plus_interval i1 (plus_interval i2 i3)).
Proof.
  kill_interval i1; kill_2intervals i2 i3.
Qed.

Lemma plus_interval_l_null : forall i,
    eq_interval (plus_interval EmptyInterval i) EmptyInterval.
Proof.
  reflexivity.
Qed.

Lemma plus_join_interval_distrib: forall i1 i2 i3,
    eq_interval (plus_interval (join_interval i1 i2) i3)
                (join_interval (plus_interval i1 i3) (plus_interval i2 i3)).
Proof.
  kill_interval i1; kill_2intervals i2 i3.
Qed.

Lemma plus_interval_cong_l : forall i i1 i2,
    eq_interval i1 i2 -> eq_interval (plus_interval i1 i) (plus_interval i2 i).
Proof.
  intros; destruct i1, i2, i;
    repeat match goal with
           | [ s : {'(_, _) : Z * Z | (_ <= _)%Z } |- _ ] => destruct s as [x ?]; destruct x as [? ?]
           end; simpl in *; f_equal; auto; try ring;
      repeat match goal with
             | [ H : (_, _) = (_, _) |- _] => inversion H; subst; try ring
             end; try inversion H; eauto.
Qed.

Lemma plus_interval_cong_r : forall i i1 i2,
    eq_interval i1 i2 -> eq_interval (plus_interval i i1) (plus_interval i i2).
Proof.
  intros; etransitivity;
  [apply plus_interval_comm | etransitivity];
  [apply plus_interval_cong_l; eauto | apply plus_interval_comm].
Qed.

Add Morphism plus_interval
  with signature eq_interval ==> eq_interval ==> eq_interval
  as plus_interval_morph.
Proof.
  intros;
  apply plus_interval_cong_l with (i:= x0) in H;
  apply plus_interval_cong_r with (i:=y) in H0;
  etransitivity; eauto.
Qed.

Lemma IntervalTheory:
  semi_ring_theory EmptyInterval (singleton_interval 0) join_interval plus_interval eq_interval.
Proof.
  split.
  - apply join_interval_l_id.
  - apply join_interval_comm.
  - symmetry; apply join_interval_assoc.
  - apply plus_interval_l_zero.
  - apply plus_interval_l_null.
  - apply plus_interval_comm.
  - symmetry; apply plus_interval_assoc.
  - apply plus_join_interval_distrib.
Qed.

Lemma interval_eq_dec: forall i1 i2 : interval, {eq_interval i1 i2} + {~(eq_interval i1 i2)}.
Proof.
  intros.
  destruct i1, i2; simpl; try (right; discriminate); try (left; reflexivity).
  destruct s, s0; simpl; repeat decide equality.
  destruct (Z.eq_dec z z0); subst; auto; right; intro H; elim n; inversion H; subst; auto.
  destruct (Z.eq_dec z z0); subst; auto; right; intro H; elim n; inversion H; subst; auto.
Qed.

Lemma IntervalMorphism:
  sring_eq_ext join_interval plus_interval eq_interval.
Proof.
  split.
  - exact join_interval_eqmor.
  - exact plus_interval_morph.
Qed.
Add Ring interval_ring : IntervalTheory (setoid eq_interval_rel IntervalMorphism).

Lemma In_interval_plus_singleton : forall z1 z2 i,
    In_interval z1 (plus_interval (singleton_interval z2) i) <->
    In_interval (z1 - z2) i.
Proof.
  kill_interval i.
Qed.

Lemma In_interval_join_intro : forall z i i', In_interval z i \/ In_interval z i' -> In_interval z (join_interval i i').
Proof.
  kill_2intervals i i'.
Qed.

Lemma join_interval_is_lub : forall i i', Incl_interval i (join_interval i i') /\ Incl_interval i' (join_interval i i').
Proof.
  autounfold with csl; intuition; kill_2intervals i i'.
Qed.

Lemma Incl_interval_empty : forall i,
    Incl_interval EmptyInterval i.
Proof.
  autounfold with csl; intros; kill_interval i.
Qed.

Lemma Incl_interval_full : forall i,
    Incl_interval i FullInterval.
Proof.
  autounfold with csl; kill_interval i.
Qed.

Definition lb (s : {'(x, y) : Z * Z | (x <= y)%Z}) : Z := let (a, _) := proj1_sig s in a.
Definition ub (s : {'(x, y) : Z * Z | (x <= y)%Z}) : Z := let (_, b) := proj1_sig s in b.

Lemma Incl_interval_bounds_ir : forall s z,
    Incl_interval (Interval s) (Above z) ->
    (z <= lb s /\ z <= ub s)%Z.
Proof.
  autounfold with csl; intros; destruct s as [[? ?] ?];
  unfold lb, ub;  simpl in *; split.
  - destruct (Z_lt_le_dec z0 z); [ | lia];
    destruct (H z0); [lia | apply Z.compare_gt_iff; auto].
  - destruct (Z_lt_le_dec z1 z); [ | lia];
    destruct (H z0); [lia | apply Z.compare_gt_iff; lia].
Qed.

Lemma Incl_interval_bounds_il : forall s z,
    Incl_interval (Interval s) (Below z) ->
    (lb s <= z /\ ub s <= z)%Z.
Proof.
  autounfold with csl; intros; destruct s as [[? ?] ?];
  unfold lb, ub; simpl in *; split.
  - destruct (Z_lt_le_dec z z0); [ | lia];
    destruct (H z0); [lia | apply Z.compare_gt_iff; auto].
  - destruct (Z_lt_le_dec z z1); [ | lia];
    destruct (H z1); [lia | apply Z.compare_gt_iff; auto].
Qed.

Lemma Incl_interval_bounds_ii : forall s1 s2,
    Incl_interval (Interval s1) (Interval s2) ->
    (lb s2 <= lb s1 /\ ub s1 <= ub s2)%Z.
Proof.
  autounfold with csl; intros;
  destruct s1 as [[? ?] ?];
  destruct s2 as [[? ?] ?];
  unfold lb, ub;   simpl in *; split.
  - destruct (Z_lt_le_dec z z1); [destruct (H z) as [? ?] | ]; lia.
  - destruct (Z_lt_le_dec z2 z0); [destruct (H z0) as [? ?] | ]; lia.
Qed.

Lemma Incl_interval_bounds_ll : forall z z0,
    Incl_interval (Below z) (Below z0) ->
    (z <= z0)%Z.
Proof.
  autounfold with csl; simpl; intros;
  destruct (Z_lt_le_dec z0 z); [destruct (H z) | ]; try lia;
  apply Z.compare_gt_iff; auto.
Qed.

Lemma Incl_interval_bounds_rr : forall z z0,
    Incl_interval (Above z) (Above z0) ->
    (z0 <= z)%Z.
Proof.
  autounfold with csl; simpl; intros;
  destruct (Z_lt_le_dec z z0);[destruct (H z) | ]; try lia;
  apply Z.compare_gt_iff; auto.
Qed.

Lemma not_Incl_interval_li : forall b s, not (Incl_interval (Below b) (Interval s)).
Proof.
  autounfold with csl;  destruct s as [[? ?] ?]; intuition; simpl in *;
  destruct (Z_lt_le_dec b z); [ destruct (H b) | destruct (Z_lt_le_dec b z0)]; try lia;
  destruct (H (z - 1)%Z); lia.
Qed.

Lemma not_Incl_interval_ri : forall a s, not (Incl_interval (Above a) (Interval s)).
Proof.
  autounfold with csl; destruct s as [[? ?] ?]; intuition; simpl in *;
  destruct (Z_lt_le_dec a z); [destruct (H a) | destruct (Z_lt_le_dec a z0)]; try lia;
  [destruct (H (z0 + 1)%Z) | destruct (H (a + 1)%Z)]; lia.
Qed.

Lemma not_Incl_interval_lr : forall a b, not (Incl_interval (Below b) (Above a)).
Proof.
  autounfold with csl; intuition; simpl in *;
  destruct (Z_lt_le_dec b a); [specialize (H (b - 1)%Z) | specialize (H (a - 1)%Z)]; lia.
Qed.

Lemma not_Incl_interval_rl : forall a b, not (Incl_interval (Above b) (Below a)).
Proof.
  autounfold with csl; intuition; simpl in *;
    destruct (Z_lt_le_dec a b); [specialize (H (b + 1)%Z) | specialize (H (a + 1)%Z)]; lia.
Qed.

Ltac easy_case :=
  repeat match goal with
  | [ i : interval |- _ ] => destruct i as [ | [[?]] | ? | ? | ]
  | [ s : {'(x, y) : Z * Z | (x <= y)%Z} |- _ ] => destruct s as [[?]]
  end; simpl in *; intuition.

Lemma not_Incl_interval_full : forall a, a <> FullInterval -> not (Incl_interval FullInterval a).
Proof.
  autounfold with csl; intros; simpl in *; destruct a; intuition.
  - contradiction (H0 0); trivial.
  - destruct s as [[? ?] ?]; simpl in *;
    destruct (H0 (z0 + 1)%Z); auto; lia.
  - destruct (H0 (z + 1)%Z); auto; apply Z.compare_gt_iff; lia.
  - destruct (H0 (z - 1)%Z); auto; apply Z.compare_gt_iff; lia.
Qed.

Lemma not_Incl_interval_empty : forall a, a <> EmptyInterval -> not (Incl_interval a EmptyInterval).
Proof.
  autounfold; intros; simpl in *; destruct a; intuition;
    try (destruct s as [[? ?] ?]); try (apply (H0 z); simpl in *; lia);
      apply (H0 0); simpl in *; lia.
Qed.

Hint Resolve not_Incl_interval_li not_Incl_interval_ri not_Incl_interval_ri not_Incl_interval_lr not_Incl_interval_rl not_Incl_interval_empty not_Incl_interval_full : csl.

Lemma plus_interval_l_monotone : forall a a' b,
    Incl_interval a a' ->
    Incl_interval (plus_interval a b) (plus_interval a' b).
Proof.
  intros; destruct a, a'; try (contradict H; progress eauto with csl);
    try (unfold Incl_interval in *; simpl in *; easy);
    try (contradict H; apply not_Incl_interval_empty; progress intuition; inversion H);
    try (contradict H; apply not_Incl_interval_full; progress intuition; inversion H).
  - apply Incl_interval_bounds_ii in H as H'; autounfold with csl; unfold lb, ub in *; easy_case.
  - apply Incl_interval_bounds_il in H as H'; autounfold with csl; unfold lb, ub in *; easy_case.
  - apply Incl_interval_bounds_ir in H as H'; autounfold with csl; unfold lb, ub in *; easy_case.
  - autounfold with csl; easy_case.
  - apply Incl_interval_bounds_ll in H as H'; autounfold with csl; unfold lb, ub in *; easy_case.
  - autounfold with csl; easy_case.
  - apply Incl_interval_bounds_rr in H as H'; autounfold with csl; unfold lb, ub in *; easy_case.
  - autounfold with csl; easy_case.
Qed.

Lemma plus_interval_r_monotone : forall a b b',
    Incl_interval b b' ->
    Incl_interval (plus_interval a b) (plus_interval a b').
Proof.
  intros; etransitivity;
  [apply Incl_interval_eq; apply plus_interval_comm | ];
  eapply plus_interval_l_monotone in H; rewrite H;
  eapply Incl_interval_eq; apply plus_interval_comm.
Qed.

Lemma plus_interval_monotone : forall a a' b b',
    Incl_interval a a' ->
    Incl_interval b b' ->
    Incl_interval (plus_interval a b) (plus_interval a' b').
Proof.
  intros; etransitivity; [apply plus_interval_l_monotone | apply plus_interval_r_monotone]; auto.
Qed.

Add Morphism plus_interval
    with signature Incl_interval ==> Incl_interval ==> Incl_interval
      as plus_interval_inclmor.
Proof.
  intros; apply plus_interval_monotone; auto.
Qed.

Lemma Incl_interval_l_monotone : forall i i1 i2,
    Incl_interval i1 i2 ->
    Incl_interval (join_interval i1 i) (join_interval i2 i).
Proof.
  intros; destruct i1, i2; try (simpl in *; easy);
    try (contradict H; apply not_Incl_interval_empty; intuition; inversion H);
    try (contradict H; apply not_Incl_interval_full;  intuition; inversion H);
    try (etransitivity; [apply Incl_interval_eq; apply join_interval_l_id | apply join_interval_is_lub]).
  - apply Incl_interval_bounds_ii in H; destruct s as [[? ?]]; destruct s0 as [[? ?]];
      autounfold with csl; unfold lb, ub in *; simpl in H; intros; kill_interval i.
  - apply Incl_interval_bounds_il in H; destruct s as [[? ?]];
      autounfold with csl; unfold lb, ub in *; simpl in H; intros; kill_interval i.
  - apply Incl_interval_bounds_ir in H; destruct s as [[? ?]];
      autounfold with csl; unfold lb, ub in *; simpl in H; intros; kill_interval i.
  - etransitivity; [apply Incl_interval_full | apply Incl_interval_eq; symmetry; apply join_interval_l_full].
  - contradict H; apply not_Incl_interval_li; intuition; inversion H.
  - apply Incl_interval_bounds_ll in H; autounfold with csl; intros; kill_interval i.
  - contradict H; apply not_Incl_interval_lr; intuition; inversion H.
  - etransitivity; [apply Incl_interval_full | apply Incl_interval_eq; symmetry; apply join_interval_l_full].
  - contradict H; apply not_Incl_interval_ri; intuition; inversion H.
  - contradict H; apply not_Incl_interval_rl; intuition; inversion H.
  - apply Incl_interval_bounds_rr in H; autounfold with csl; intros; kill_interval i.
  - etransitivity; [apply Incl_interval_full | apply Incl_interval_eq; symmetry; apply join_interval_l_full].
Qed.

Lemma Incl_interval_r_monotone : forall i i1 i2,
    Incl_interval i1 i2 ->
    Incl_interval (join_interval i i1) (join_interval i i2).
Proof.
  intros; eapply Incl_interval_l_monotone in H; etransitivity;
    [ apply Incl_interval_eq; apply join_interval_comm | ];
    etransitivity; [exact H | ];
      apply Incl_interval_eq;
      apply join_interval_comm.
Qed.

Lemma Incl_interval_monotone : forall i1 i2 i3 i4,
    Incl_interval i1 i2 /\ Incl_interval i3 i4 ->
    Incl_interval (join_interval i1 i3) (join_interval i2 i4).
Proof.
  intros; destruct H; etransitivity;
  [apply Incl_interval_l_monotone | apply Incl_interval_r_monotone]; auto.
Qed.

Add Morphism join_interval
    with signature Incl_interval ==> Incl_interval ==> Incl_interval
      as join_interval_inclmor.
Proof.
  intros; apply Incl_interval_monotone; auto.
Qed.

Program Instance interval_Lattice : Lattice interval :=
  {
    top := FullInterval;
    bot := EmptyInterval;
    Incl := Incl_interval;
    Incl_bot := Incl_interval_empty;
    Incl_top := Incl_interval_full;
    Incl_trans := Incl_interval_trans;
    join := join_interval;
    join_correct := join_interval_is_lub
  }.
Next Obligation.
  apply Incl_interval_refl.
Defined.
Next Obligation.
  apply Incl_interval_monotone; auto.
Defined.

Lemma way3_thing : forall i1 i2,
    In_interval 0
                  (plus_interval (join_interval (singleton_interval 0) i1)
                                 (join_interval (singleton_interval 0) i2)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma way3_thing' : forall i1 i2,
    Incl_interval i1 (plus_interval (join_interval (singleton_interval 0) i1)
                                    (join_interval (singleton_interval 0) i2)).
Proof.
  autounfold with csl; intros; kill_2intervals i1 i2.
Qed.

Lemma way3_thing'' : forall i1 i2,
    Incl_interval i2 (plus_interval (join_interval (singleton_interval 0) i1)
                                    (join_interval (singleton_interval 0) i2)).
Proof.
  autounfold with csl; intros; kill_2intervals i1 i2.
Qed.

Lemma In_interval_zero_plus : forall z i1 i2,
    In_interval z i1 -> In_interval (- z) i2 -> In_interval 0 (plus_interval i1 i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma In_interval_zero_plus' : forall z i1 i2,
    In_interval z i1 -> In_interval (- z) i2 ->
    In_interval 0 (plus_interval (join_interval (singleton_interval 0) i1) i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma In_interval_zero_plus'' : forall z i1 i2,
    In_interval z i1 -> In_interval (- z) i2 ->
    In_interval 0 (plus_interval i1 (join_interval (singleton_interval 0) i2)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact1 : forall r i1 i2,
    In_interval (- r) i1 -> In_interval r i2 ->
    In_interval 0 (plus_interval (join_interval (singleton_interval 0) i1) i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact2 : forall r i1 i2,
    In_interval (- r) i1 -> In_interval r i2 ->
    In_interval r (plus_interval (join_interval (singleton_interval 0) i1) i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact3 : forall r i1 i2,
    In_interval (- r) i1 -> In_interval 0 i2 ->
    In_interval (- r) (plus_interval (join_interval (singleton_interval 0) i1) i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact4 : forall r i1 i2,
    In_interval (- r) i1 ->
    In_interval r i2 ->
    In_interval r (plus_interval (join_interval (singleton_interval 0) i1) i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact1': forall r i1 i2,
    In_interval (- r) i1 -> In_interval r i2 ->
    In_interval 0 (plus_interval i2 (join_interval (singleton_interval 0) i1)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact2' : forall r i1 i2,
    In_interval (- r) i1 -> In_interval r i2 ->
    In_interval r (plus_interval i2 (join_interval (singleton_interval 0) i1)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact3' : forall r i1 i2,
    In_interval (- r) i1 -> In_interval 0 i2 ->
    In_interval (- r) (plus_interval i2 (join_interval (singleton_interval 0) i1)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact4' : forall r i1 i2,
    In_interval (- r) i1 ->
    In_interval r i2 ->
    In_interval r (plus_interval i2 (join_interval (singleton_interval 0) i1)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact5 : forall r i1 i2,
    In_interval 0 i1 ->
    In_interval r i2 ->
    In_interval r (plus_interval i1 (join_interval (singleton_interval 0) i2)).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma In_interval_plus_zero : forall i1 i2,
    In_interval 0 i1 -> In_interval 0 i2 -> In_interval 0 (plus_interval i1 i2).
Proof. kill_2intervals i1 i2. Qed.

Lemma In_interval_join_intro1 : forall z i,
    In_interval z (join_interval (singleton_interval z) i).
Proof. kill_interval i. Qed.

Lemma In_interval_join_intro2 : forall z i,
  In_interval z (join_interval i (singleton_interval z)).
Proof. kill_interval i. Qed.

Lemma In_interval_plus : forall r i1 i2,
  In_interval (- r) i1 -> In_interval r i2 -> In_interval 0 (plus_interval i1 i2).
Proof. kill_2intervals i1 i2. Qed.

Lemma fact6 : forall r i1 i2,
    In_interval (- r) i1 ->
    In_interval 0 i2 ->
    In_interval (- r) (plus_interval i1 i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma fact7 : forall r i1 i2,
    In_interval 0 i1 ->
    In_interval r i2 ->
    In_interval r (plus_interval i1 i2).
Proof.
  kill_2intervals i1 i2.
Qed.

Lemma plus_interval_cong_assoc : forall i i1 i2 i3,
    eq_interval i1 (plus_interval i2 i3) ->
    eq_interval (plus_interval i i1)
                (plus_interval (plus_interval i i2) i3).
Proof.
  intros; rewrite H; ring.
Qed.

Lemma plus_interval_cong_assoc_comm : forall i i1 i2 i3,
    eq_interval i1 (plus_interval i2 i3) ->
    eq_interval (plus_interval i i1)
                (plus_interval i2 (plus_interval i i3)).
Proof.
  intros; rewrite H; ring.
Qed.

Ltac interval_solver tac:=
  tac;
  match goal with
  | [ H : Incl_interval (singleton_interval _) _ |- _ ] =>
    apply In_interval_Incl_interval in H; interval_solver tac
  | |- Incl_interval (singleton_interval _ ) _ =>
    apply In_interval_Incl_interval; interval_solver tac
  | _ => idtac
  end; eauto with csl.
