Require Import List ListSet Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.AbstractValues.

Set Implicit Arguments.

(* Result of the analysis *)
Definition abstract_agent_pair := (abstract_value Agent * abstract_value Agent)%type.

Definition dec_eq_aap : forall (x y : abstract_agent_pair), {x = y} + {x <> y}.
  repeat decide equality.
Defined.

(* Abstract cartesian product from two abstract sets in the value map
   to the resulting lattice. *)
Definition cross (s1 s2 : abstract_set Agent) : set abstract_agent_pair :=
  match s1, s2 with
  | FullSet, FullSet => singleton dec_eq_aap (AnyValue, AnyValue)
  | ActualSet s, ActualSet s' =>
    set_map dec_eq_aap (fun v => match v with (v1, v2) => (ActualValue v1, ActualValue v2) end) (set_prod s s')
  | FullSet, ActualSet s => set_map dec_eq_aap (fun v => (AnyValue, ActualValue v)) s
  | ActualSet s, FullSet => set_map dec_eq_aap (fun v => (ActualValue v, AnyValue)) s
  end.


(* Helpers for representation function *)
Program Definition ev_to_aap e : abstract_agent_pair :=
  match e with
  | Event a1 a2 _ _ => (ActualValue a1, ActualValue a2)
  end.

Definition β_participation_folder (e : event) (s : set abstract_agent_pair) : set abstract_agent_pair :=
  set_add dec_eq_aap (ev_to_aap e) s.

(* Representation function *)
Definition β_participation (t : trace) : set (abstract_agent_pair) :=
  fold_right β_participation_folder (empty_set _) t.

Hint Resolve set_union_elim set_union_intro set_union_intro1 set_union_intro2 set_add_elim set_add_intro set_add_intro1 set_add_intro2 : csl.
Hint Unfold β_participation_folder : csl.

(* Proof of homomorphisms for β_participation *)
Lemma β_participation_app_union : forall (t1 t2 t3 : trace) a,
    appends t1 t2 t3 -> set_In a (β_participation t3) ->
    set_In a (set_union dec_eq_aap (β_participation t1) (β_participation t2)).
Proof.
  intros.
  induction H; simpl in *; auto with csl.
  unfold β_participation_folder in *.
  simpl in *.

  apply set_add_elim in H0.
  destruct H0; auto with csl.
  apply IHappends in H0;
  apply set_union_elim in H0.
  destruct H0; auto with csl.
Qed.

Lemma β_participation_interleave_union : forall (t1 t2 t3 : trace) a,
    interleave t1 t2 t3 -> set_In a (β_participation t3) ->
    set_In a (set_union dec_eq_aap (β_participation t1) (β_participation t2)).
Proof.
  intros.
  induction H; auto with csl; simpl in *;
    unfold β_participation_folder in *;
    apply set_add_elim in H0;
    destruct H0; auto with csl;
      apply IHinterleave in H0;
      apply set_union_elim in H0;
      intuition.
Qed.

(* What it means to be included when we have AnyValue.
   We can just reuse the definition for set_union.
 *)
Definition aap_set_In (aap : abstract_agent_pair) (s : set abstract_agent_pair) : Prop :=
  match aap with
  | (AnyValue, AnyValue) => set_In aap s
  | (ActualValue v, AnyValue) => set_In aap s \/ set_In (AnyValue, AnyValue) s
  | (AnyValue, ActualValue v) => set_In aap s \/ set_In (AnyValue, AnyValue) s
  | (ActualValue v, ActualValue v') =>
    set_In aap s \/ set_In (AnyValue, AnyValue) s \/ set_In (AnyValue, ActualValue v') s \/ set_In (ActualValue v, AnyValue) s
  end.

Definition aap_set_Incl (s1 s2 : set abstract_agent_pair) :=
  forall a, aap_set_In a s1 -> aap_set_In a s2.

Lemma aap_set_union_intro : forall a s1 s2,
    aap_set_In a s1 \/ aap_set_In a s2 -> aap_set_In a (set_union dec_eq_aap s1 s2).
Proof.
  intros; unfold aap_set_In in *.
  destruct a.
  destruct a, a0; simpl; auto;
  repeat (destruct H; simpl; auto with csl).
Qed.

Hint Resolve set_union_elim : csl.

Lemma aap_set_union_elim : forall a s1 s2,
    aap_set_In a (set_union dec_eq_aap s1 s2) -> aap_set_In a s1 \/ aap_set_In a s2.
Proof.
  intros; unfold aap_set_In in *.

  destruct a.
  destruct a, a0; simpl; intuition;
    try match goal with
    | [ H : set_In _ (set_union _ _ _) |- _ ] => apply set_union_elim in H; intuition
    end.
Qed.

Lemma aap_set_add_elim:
  forall a h l,
    aap_set_In a (set_add dec_eq_aap (ev_to_aap h) l) -> a = ev_to_aap h \/ aap_set_In a l.
Proof.
  intros; unfold aap_set_In in *.
  destruct h;
  destruct a as [a a0];
  destruct a, a0; simpl in *; intuition;
    try match goal with
    | [ H : set_In _ (set_add _ _ _) |- _ ] => apply set_add_elim in H; intuition
    end;
  try inversion H; inversion H0.
Qed.

Lemma aap_set_In_weak : forall a s,
    set_In a s -> aap_set_In a s.
Proof.
  intros.
  destruct a;
  destruct a, a0; simpl; eauto.
Qed.

Hint Unfold aap_set_Incl : csl.
Hint Resolve set_add_elim set_add_intro aap_set_union_intro : csl.
Hint Unfold β_participation_folder : csl.

(* True since β_participation does not add any ⊤ values *)
Lemma aap_set_β_participation : forall a t,
    aap_set_In a (β_participation t) ->
    set_In a (β_participation t).
Proof.
  intros.
  induction t; [ destruct a; destruct a, a0; simpl in *; intuition
                | repeat (match goal with
                         | [ e : event |- _ ] => destruct e
                         | [ a : abstract_agent_pair |- _ ] => destruct a
                         | [ a : abstract_value _ |- _ ] => destruct a
                         | [ H : set_In _ (β_participation_folder _ _) |- _ ] =>
                           unfold β_participation_folder in *; apply set_add_elim in H; destruct H; [ inversion H | ]
                         end; simpl in *; intuition; eauto) ].
Qed.

Hint Resolve set_add_intro set_add_elim : csl.

(* Proofs about being included in the cartesian product when we have abstract values.
   We start by some nasty lemmas.
 *)
Lemma set_In_prod_aux : forall (a1 a2 : tyDenote Agent) s,
        set_In a2 s -> set_In (ActualValue a1, ActualValue a2) (set_map dec_eq_aap (fun v => (ActualValue a1, ActualValue v)) s).
Proof.
  intros.
  induction s; auto.
  simpl in *.
  destruct H; subst; auto with csl.
Qed.

Lemma set_In_map_map : forall v (a : tyDenote Agent) (l : set (tyDenote Agent)),
    set_In v (set_map dec_eq_aap (fun y => (ActualValue a, ActualValue y)) l) ->
    set_In v (set_map dec_eq_aap (fun v : tyDenote Agent * tyDenote Agent =>
                                    let (v1, v2) := v in (ActualValue v1, ActualValue v2))
                      (map (fun y : tyDenote Agent => (a, y)) l)).
Proof.
  intros.
  induction l; auto.
  simpl in *.
  apply set_add_elim in H as [? | ?]; auto with csl.
Qed.

Lemma set_In_set_map_app :  forall (v : abstract_agent_pair) s1 s2 (f : (tyDenote Agent * tyDenote Agent) -> abstract_agent_pair),
    set_In v (set_map dec_eq_aap f s1) \/ set_In v (set_map dec_eq_aap f s2) ->
    set_In v (set_map dec_eq_aap f (s1 ++ s2)%list).
Proof.
  intros.
  induction s1; simpl in *; intuition.
  apply set_add_elim in H0 as [? | ?]; auto with csl.
Qed.

Hint Resolve set_In_set_map_app set_In_prod_aux set_In_map_map : csl.

Lemma set_In_prod : forall (a1 a2 : tyDenote Agent) s1 s2,
    set_In a1 s1 ->
    set_In a2 s2 ->
    set_In (ActualValue a1, ActualValue a2)
           (set_map dec_eq_aap (fun v => match v with (v1, v2) => (ActualValue v1, ActualValue v2) end) (set_prod s1 s2)).
Proof.
  intros.
  generalize dependent s2.
  induction s1; intros; simpl in *; eauto.
  destruct H; subst; eauto with csl.
Qed.


(* This is the actual lemma we want to prove *)
Lemma abstract_set_In_cross : forall a1 a2 s1 s2,
  abstract_set_In a1 s1 /\ abstract_set_In a2 s2 ->
  aap_set_In (ActualValue a1, ActualValue a2) (cross s1 s2).
Proof.
  intros.
  let t :=
      induction s;
        inversion H1;
        simpl in *; intuition; subst; auto with csl
    in
      destruct s1, s2; simpl in *; intuition; [ t | t | ].
    left.
    apply set_In_prod; auto with csl.
Qed.

(* Ugly ugly.. *)
Lemma set_In_L_Any_Actual_False: forall a s,
  set_In (a, AnyValue) (set_map dec_eq_aap (fun v => (AnyValue, ActualValue v)) s) -> False.
  intros; induction s; simpl in *;  [ | apply IHs]; auto.
  apply set_add_elim in H; destruct H; [inversion H | assumption].
Qed.

Lemma set_In_Any_Actual_False : forall a s,
    set_In (a, AnyValue)
           (set_map dec_eq_aap (fun v : tyDenote Agent * tyDenote Agent => let (v1, v2) := v in (ActualValue v1, ActualValue v2)) s) -> False.
  intros; induction s; simpl in *;  [ | apply IHs]; auto.
  apply set_add_elim in H; destruct H; [destruct a0; inversion H | assumption].
Qed.

Lemma set_In_Any_Actual_False' : forall a s,
    set_In (AnyValue, a)
           (set_map dec_eq_aap (fun v : tyDenote Agent * tyDenote Agent => let (v1, v2) := v in (ActualValue v1, ActualValue v2)) s) -> False.
  intros; induction s; simpl in *;  [ | apply IHs]; auto.
  apply set_add_elim in H; destruct H; [destruct a0; inversion H | assumption].
Qed.

Lemma set_In_R_Any_Actual_False: forall a s,
  set_In (AnyValue, a) (set_map dec_eq_aap (fun v => (ActualValue v, AnyValue)) s) -> False.
  intros; induction s; simpl in *;  [ | apply IHs]; auto.
  apply set_add_elim in H; destruct H; [inversion H | assumption].
Qed.

Lemma set_In_prod_L:
  forall (a1 a2 : tyDenote Agent) (s1 s2 : set (tyDenote Agent)),
  set_In a1 s1 ->
  set_In (ActualValue a1, AnyValue)
         (set_map dec_eq_aap (fun v => (ActualValue v, AnyValue)) s1).
  intros; induction s1; intros; simpl in *; [ | destruct H; subst ]; auto with csl.
Qed.

Lemma set_In_prod_R:
  forall (a1 a2 : tyDenote Agent) (s1 s2 : set (tyDenote Agent)),
  set_In a1 s1 ->
  set_In (AnyValue, ActualValue a1)
    (set_map dec_eq_aap (fun v => (AnyValue, ActualValue v)) s1).
  intros; induction s1; intros; simpl in *; [ | destruct H; subst ]; auto with csl.
Qed.

Lemma set_In_map_app:
  forall A B (l1 l2 : set A) (f : A -> B) dec (v : B),
    set_In v (set_map dec f (l1 ++ l2)) -> @set_In B v (set_map dec f l1) \/ set_In v (set_map dec f l2).
  induction l1; intros; simpl in *; auto.
  apply set_add_elim in H; destruct H.
  + subst; auto with csl.
  + destruct IHl1 with l2 f dec v; auto with csl.
Qed.

Lemma set_In_map_map_imp:
  forall A B (l : set A) (f : (A * A) ->  B * B) (g : A -> A * A) dec (v : B * B),
    set_In v (set_map dec f (map g l)) -> exists (w : A), f (g w) = v /\ set_In w l.
  induction l; intros; simpl in *; [contradiction | ].
  apply set_add_elim in H; destruct H.
  + subst; exists a; split; auto.
  + destruct IHl with f g dec v; auto.
    destruct H0; exists x; split; auto.
Qed.

Lemma set_In_Any_Actual:
  forall s s0 t a,
    (forall v, set_In v s -> set_In v s0) ->
    set_In (a, ActualValue t) (set_map dec_eq_aap (fun v => (AnyValue, ActualValue v)) s) ->
    set_In t s0.
  intros. induction s; intros; simpl in *; [contradiction | apply set_add_elim in H0; destruct H0].
  + inversion H0; subst; auto.
  + apply IHs; auto.
Qed.

Lemma set_In_Actual_Any:
  forall s s0 t a,
    (forall v, set_In v s -> set_In v s0) ->
    set_In (ActualValue t, a) (set_map dec_eq_aap (fun v => (ActualValue v, AnyValue)) s) ->
    set_In t s0.
  intros. induction s; intros; simpl in *; [contradiction | apply set_add_elim in H0; destruct H0].
  + inversion H0; subst; auto.
  + apply IHs; auto.
Qed.

Lemma set_In_Actual_Actual:
  forall s s' s1 t0 t1,
    (forall v, set_In v s' -> set_In v s1) ->
    set_In (ActualValue t0, ActualValue t1)
           (set_map dec_eq_aap
                    (fun v : tyDenote Agent * tyDenote Agent => let (v1, v2) := v in (ActualValue v1, ActualValue v2))
                    (set_prod s s')) ->
    set_In t1 s1.
  induction s; simpl; intros; [ contradiction | ].
  apply set_In_map_app in H0; destruct H0.
  + apply set_In_map_map_imp in H0.
    destruct H0 as [w [H0 H1 ]].
    inversion H0; subst; auto.
  + eapply IHs; eauto.
Qed.

Lemma set_In_Actual_Actual':
  forall s s' s0 t0 t1,
    (forall v, set_In v s -> set_In v s0) ->
    set_In (ActualValue t0, ActualValue t1)
           (set_map dec_eq_aap
                    (fun v : tyDenote Agent * tyDenote Agent => let (v1, v2) := v in (ActualValue v1, ActualValue v2))
                    (set_prod s s')) ->
    set_In t0 s0.
  induction s; simpl; intros; [ contradiction | ].
  apply set_In_map_app in H0; destruct H0.
  + apply set_In_map_map_imp in H0.
    destruct H0 as [w [H0 H1 ]].
    inversion H0; subst; auto.
  + eapply IHs; eauto.
Qed.

Hint Resolve set_In_Any_Actual set_In_Actual_Any set_In_Actual_Actual set_In_Actual_Actual' : csl.

Ltac try_contradict_set_In :=
  match goal with
  | [ H : set_In _ _ |- _ ] => try (apply set_In_L_Any_Actual_False in H; contradiction);
                          try (apply set_In_R_Any_Actual_False in H; contradiction);
                          try (apply set_In_Any_Actual_False in H; contradiction);
                          try (apply set_In_Any_Actual_False' in H; contradiction)
  end.

Lemma aap_set_in_cross_help : forall a s1 s2 s1' s2',
    abstract_set_Incl s1 s1' ->
    abstract_set_Incl s2 s2' ->
    aap_set_In a (cross s1 s2) ->
    aap_set_In a (cross s1' s2').
Proof.
  intros; destruct a; destruct a, a0; simpl in *.
  - destruct s1, s2; simpl in *; intuition; destruct s1', s2'; simpl in *; intuition; try_contradict_set_In.
  - destruct s1', s2'; simpl in *; intuition; destruct s1, s2; simpl in *; intuition; try_contradict_set_In.
    + left. apply set_In_prod_R; eauto with csl.
  - destruct H1; destruct s1', s2'; simpl in *;
      intuition; destruct s1, s2; simpl in *; intuition; try_contradict_set_In.
    + left. apply set_In_prod_L; eauto with csl.
  - (* slow! *) destruct H1; [ | destruct H1]; destruct s1', s2'; simpl in *; intuition;
    destruct s1, s2; simpl in *; intuition; try_contradict_set_In.
    + right. right. left. apply set_In_prod_R; eauto with csl.
    + right. right. left. apply set_In_prod_R; eauto with csl.
    + right. right. right. apply set_In_prod_L; eauto with csl.
    + right. right. right. apply set_In_prod_L; eauto with csl.
    + left. apply set_In_prod; eauto with csl.
    + right. right. left. apply set_In_prod_R; eauto with csl.
    + right. right. right. apply set_In_prod_L; eauto with csl.
Qed.

Lemma aap_set_In_cross : forall Δ (m m' : value_map (new_var Δ)) a,
    value_map_Incl m m' ->
    aap_set_In a (cross (hget m HFirst) (hget m (HNext HFirst))) ->
    aap_set_In a (cross (hget m' HFirst) (hget m' (HNext HFirst))).
Proof.
  intros.

  dependent destruction m; dependent destruction m.
  dependent destruction m'; dependent destruction m'.

  remember (H Agent HFirst).
  remember (H Agent (HNext HFirst)).
  eapply aap_set_in_cross_help; eauto.
Qed.
