(** printing τ $\tau$ #τ# *)
(** printing τ' $\tau'$ #τ'# *)
(** printing τ1 $\tau_1$ #τ1# *)
(** printing τ2 $\tau_2$ #τ2# *)
(** printing Δ $\Delta$ #Δ# *)
(** printing Δ' $\Delta'$ #Δ'# *)
(** printing δ $\delta$ #δ# *)
(** printing δ' $\delta'$ #δ'# *)
(** printing Γ $\Gamma$ #Γ# *)
Require Import List Uniset Bool String Omega Program.

Import ListNotations.

Require Import CSL.Syntax.
Require Import CSL.Definitions.
Require Import CSL.Satisfaction.
Require Import CSL.Exp.
Require Import CSL.HList.
Require Import CSL.Tactics.

Set Implicit Arguments.

(* Computes all possible pairs that interleaves to the list l *)
Fixpoint interleavings (T : Set) (l : list T) : list (list T * list T) :=
  match l with
  | nil => [(nil, nil)]
  | x :: xs => flat_map (fun i => match i with
                             | (i1, i2) => [(x :: i1, i2); (i1, x :: i2)]
                             end) (interleavings xs)
  end.

(* Computes all possible pairs that append to the list l *)
Definition appendings (T : Set) (l : list T) : list (list T * list T) :=
  map (fun x => (firstn x l, skipn x l)) (List.seq 0 (S (List.length l))).

Fixpoint contractDenote (n : nat) Γ Δ (c : contract Γ Δ) (D : template_env Γ) (δ : env Δ) : uniset trace :=
  match n with
  | O => Emptyset _
  | S n =>
    match c with
    | success _ _     => Singleton (fun x y => x = y) trace_dec_eq nil
    | failure _ _     => Emptyset _
    | tapp ae f       => contractDenote n (hget D f) D (arg_envDenote δ ae)
    | transfer p c    => Charac (fun t =>
                                  match t with
                                  | e :: ts =>
                                    let new_env := addEvent e δ  in
                                    let tc := contractDenote n c D new_env in
                                    charac tc ts && expDenote p new_env
                                  | _ => false                           
                                  end)
    | alternate c1 c2 => union (contractDenote n c1 D δ) (contractDenote n c2 D δ)
    | sequence c1 c2  => Charac (fun t =>
                                  let tc1 := contractDenote n c1 D δ in
                                  let tc2 := contractDenote n c2 D δ in
                                  existsb (fun pt =>
                                             match pt with
                                             | (t1, t2) => charac tc1 t1 && charac tc2 t2
                                             end) (appendings t))
    | concurrent c1 c2 => Charac (fun t =>
                                   let tc1 := contractDenote n c1 D δ in
                                   let tc2 := contractDenote n c2 D δ in
                                   existsb (fun pt =>
                                              match pt with
                                              | (t1, t2) => charac tc1 t1 && charac tc2 t2
                                              end) (interleavings t))
    end end.

Lemma leb_compl : forall a b : bool, 
    (a = true -> b = true) -> Bool.leb a b.
Proof. intros. destruct a; intuition. Qed.

Lemma leb_sound : forall a b : bool, 
    Bool.leb a b -> (a = true -> b = true).
Proof. intros; rewrite H0 in H; auto. Qed.

Theorem leb_correct : forall a b : bool, 
    (a = true -> b = true) <-> Bool.leb a b.
Proof.
  intros; split; [ apply leb_compl | apply leb_sound ].
Qed.

Lemma incl_In : forall T (a b : uniset T),
    incl a b <-> (forall x, In a x -> In b x).
Proof.
  unfold In, incl in *; intros; split; intros;
    eapply leb_correct;
    eauto.
Qed.

(* Strong definition of contract denotation *)
Definition denotes Γ Δ (δ : env Δ) (D : template_env Γ) c t :=
  exists n, forall m, n <= m -> In (contractDenote m c D δ) t.

(* Weaker definition of contract denotation *)
Definition denotes' Γ Δ (δ : env Δ) (D : template_env Γ) c t :=
  exists n, In (contractDenote n c D δ) t.

Lemma denotes'_helper: forall n Γ Δ (δ: env Δ) (D: template_env Γ) c t,
    In (contractDenote n c D δ) t ->
    forall m, m >= n -> In (contractDenote m c D δ) t.
Proof.
  induction n; [intros; inversion H | induction c]; intros;
    destruct H0; auto; simpl; simpl in H; unfold In in *; simpl in H; simpl;
      try rewrite orb_true_iff in *; try rewrite andb_true_iff in *.
  + apply IHn; auto; omega.
  + destruct t. inversion H. rewrite andb_true_iff in *.
    destruct H; split; auto.
  + destruct H; [left | right]; auto.
  + destruct H; [left | right].
    * destruct H; split; auto.
    * apply existsb_exists in H; apply existsb_exists. destruct H. exists x.
      destruct H. destruct x. split; auto.
      rewrite andb_true_iff in *; destruct H1; split; auto.
  + apply existsb_exists in H; apply existsb_exists. destruct H. exists x.
    destruct H. destruct x. split; auto.
    rewrite andb_true_iff in *; destruct H1; split; auto.
Qed.

Hint Resolve denotes'_helper : csl.
Lemma contractDenoteContinuous : forall n Γ Δ (c : contract Γ Δ) (D : template_env Γ) (δ : env Δ),
    incl (contractDenote n c D δ) (contractDenote (S n) c D δ).
Proof.
  intros; apply incl_In; eauto with csl.
Qed.

Lemma denotes_denotes'_Equiv: forall Γ Δ (δ : env Δ) (D : template_env Γ) c t,
    denotes δ D c t <-> denotes' δ D c t.
Proof.
  unfold denotes; unfold denotes'; split; intros;
  destruct H; exists x; [ | apply denotes'_helper]; auto.
Qed.

Hint Unfold In : csl.

Lemma skipn_length : forall T (t1 : list T), skipn (List.length t1) t1 = [].
Proof.
  intros; induction t1; auto.
Qed.

Lemma skipn_append : forall T (t1 t2 : list T),
    skipn (List.length t1) (t1 ++ t2) = t2.
Proof.
  intros; induction t1; auto.
Qed.

Lemma skipn_s : forall T n e (t : list T), skipn (S n) (e :: t) = skipn n t.
Proof.
  intros; induction t; auto.
Qed.

Hint Constructors appends : csl.

Lemma app_appends : forall (T : Set) (t t1 t2 : list T), t = app t1 t2 -> appends t1 t2 t.
Proof.
  intros.
  generalize dependent t.
  induction t1; intros; simpl in *; subst; auto with csl.
Qed.

Lemma appends_app : forall (T : Set) (t t1 t2 : list T), appends t1 t2 t -> t = app t1 t2.
Proof.
  intros.
  induction H; subst; auto.
Qed.

Lemma appends_firstn_skipn : forall (T : Set) x0 (t : list T), appends (firstn x0 t) (skipn x0 t) t.
Proof.
  eauto.
  intros.
  apply app_appends.
  now rewrite firstn_skipn.
Qed.

Hint Constructors interleave: csl.

Lemma interleavings_interleave : forall (T : Set) (t t1 t2 : list T),
    List.In (t1, t2) (interleavings t) -> interleave t1 t2 t.
  intros.
  generalize dependent t1.
  generalize dependent t2.
  induction t; intros; simpl in *; intuition.
  inv H0; auto with csl.
  rewrite in_flat_map in H.
  destruct H as [(x1, x2) [H1 H2]].
  inv H2; inv H; auto with csl.
  inv H0; auto with csl.
  inv H0.
Qed.

Lemma interleave_interleavings : forall (T : Set) (t t1 t2 : list T),
    interleave t1 t2 t -> List.In (t1, t2) (interleavings t).
  intros.
  generalize dependent t1.
  generalize dependent t2.
  induction t; intros; simpl in *.
  inv H; left; f_equal.
  apply in_flat_map.
  inv H.
  * exists ([], t); split; simpl; auto with csl.
  * exists (t, []); split; simpl; auto with csl.
  * exists (t0, t2); split; simpl; auto with csl.
  * exists (t1, t3); split; simpl; auto with csl.
Qed.

Hint Constructors csat : csl.
Hint Resolve appends_firstn_skipn interleavings_interleave: csl.

Lemma csat_denotes' : forall Γ Δ (δ : env Δ) (D : template_env Γ) t c,
    denotes' δ D c t -> csat δ D t c.
Proof.
  intros; unfold denotes' in *; unfold In in *; destruct H.
  (* Very general induction hypothesis. *)
  generalize dependent c.
  generalize dependent t.
  generalize dependent Δ.
  (* Induction on the number of iterations of the interpreter *)
  induction x; intros; [inversion H | ].

  (* Case analysis on the syntax, when the interpreter runs *)
  destruct c; simpl_hyp; intuition; eauto with csl.

  * (* success *)
    destruct t; intuition.

  * (* transfer *)
    destruct t; intuition.
    simpl_hyp; auto with csl.

  * (* sequence nonempty left *)
    destruct x0; simpl_hyp.
    inv H; eauto with csl.

  * (* concurrent *)
    destruct x0; simpl_hyp; eauto with csl.
Qed.

Lemma plus1minusS : forall n, n + 1 - (S n) = 0.
  intros.
  omega.
Qed.

Ltac more_iter x :=
  exists x; intros; intuition; match goal with
  | [ H : (_ <= ?x) |- _ ] => destruct x; [inversion H | ]; auto
  end.

Hint Resolve appends_app : csl. 

Hint Extern 1 (_ <= _) => abstract omega : csl.


Lemma csat_denotes'': forall Γ Δ (δ : env Δ) (D : template_env Γ) t c,
    csat δ D t c -> denotes δ D c t.
Proof.
  intros.
  induction H; intros; unfold denotes in *; unfold In in *; simpl in *.

  (* success *)
  more_iter 1.

  (* tapp *)
  destruct IHcsat.
  more_iter (S x); eauto with csl.

  (* transfer *)
  destruct IHcsat.
  more_iter (S x).
  simpl.
  intuition.

  (* alternate left *)
  destruct IHcsat.
  more_iter (S x).
  simpl_goal.
  left; eauto with csl.

  (* alternate right *)
  destruct IHcsat.
  more_iter (S x).
  simpl_goal.
  right; eauto with csl.

  (* sequence *)
  destruct IHcsat1.
  destruct IHcsat2.
  more_iter (S (x + x0)).
  simpl_goal.
  rewrite existsb_exists.
  destruct t1.
  left.
  simpl_goal.
  eauto with csl.
  inv H1; eauto with csl.
  right.
  exists (e :: t1, t2).
  simpl_goal.
  apply appends_app in H1.
  exists (List.length t1 + 1).
  subst.
  simpl_goal.
  f_equal.
  rewrite app_comm_cons, firstn_app, plus1minusS, app_nil_r, Nat.add_1_r, firstn_cons; f_equal.
  apply firstn_all.
  rewrite Nat.add_1_r, skipn_s, skipn_append; reflexivity.
  rewrite in_seq, app_length; omega.
  eauto with csl.
  eauto with csl.

  (* concurrent *)
  destruct IHcsat1.
  destruct IHcsat2.
  more_iter (S (x + x0)).
  simpl_goal.
  exists (t1, t2).
  split.
  apply interleave_interleavings.
  assumption.
  simpl_goal; eauto with csl.
Qed.

Theorem csat_denotes : forall Γ Δ (δ : env Δ) (D : template_env Γ) t c,
    csat δ D t c <-> denotes δ D c t.
  intuition.
  apply csat_denotes''. assumption.
  apply csat_denotes'.
  apply denotes_denotes'_Equiv.
  auto.
Qed.
