Require Ascii.

Require Import Program Bool Sumbool FunctionalExtensionality.
Require Import Coq.Strings.String.
Require Import CSL.Definitions.
Require Import CSL.Syntax.
Require Import CSL.HList.

Set Implicit Arguments.

Open Scope string_scope.



Open Scope list_scope.

Fixpoint resourceAdd (xs : tyDenote Resource) (ys: tyDenote Resource) : tyDenote Resource :=
  match xs with
  | nil => ys
  | (k1, v1) ::  xs' =>
    match List.find (fun x => String.eqb (fst x) k1) ys with
    | None => (k1, v1) :: resourceAdd xs' ys
    | Some (_, v2) => (k1, v1 + v2) :: resourceAdd xs' ys
    end
  end.

Definition equalTy {τ : ty}: tyDenote τ -> tyDenote τ -> tyDenote Bool :=
  match τ return tyDenote τ -> tyDenote τ -> tyDenote Bool with
  | Agent => String.eqb
  | Resource => resourceEqb
  | Timestamp => Nat.eqb
  | Bool => Bool.eqb
  | Scale => Nat.eqb
  end.

Lemma equalTy_refl : forall τ (e : tyDenote τ),
    equalTy e e = true.
Proof.
  destruct τ; simpl; intros;
    [apply eqb_eq | apply resourceEqb_eq | apply PeanoNat.Nat.eqb_eq
     | rewrite Bool.eqb_true_iff  | apply PeanoNat.Nat.eqb_eq]; auto.
Qed.
  
Definition lessTy {τ} (t : Comparable τ) : tyDenote τ -> tyDenote τ -> tyDenote Bool :=
  match t in Comparable τ return tyDenote τ -> tyDenote τ -> tyDenote Bool with
  | ResourceCmp => resourceLeb
  | TimestampCmp => Nat.leb
  | ScaleCmp => Nat.leb
  end.

Definition plusTy {τ} (t : Addable τ) : tyDenote τ -> tyDenote τ -> tyDenote τ :=
  match t in Addable τ return tyDenote τ -> tyDenote τ -> tyDenote τ with
  | ResourceAdd => resourceAdd
  | ScaleAdd => Nat.add
  end.

Definition multDenote (s : tyDenote Scale) (r: tyDenote Resource) : tyDenote Resource :=
  List.map (fun kv => (fst kv, s * snd kv)) r.

Fixpoint binopDenote τ1 τ2 τ3 (o : binop τ1 τ2 τ3) : tyDenote τ1 -> tyDenote τ2 -> tyDenote τ3 :=
  match o in binop τ1 τ2 τ3 return tyDenote τ1 -> tyDenote τ2 -> tyDenote τ3 with
  | And => andb
  | Or => orb
  | Mult => multDenote
  | Plus pf => plusTy pf
  | Eq _ => equalTy
  | Le pf => lessTy pf
  end.

Fixpoint unopDenote τ1 τ2 (o : unop τ1 τ2) : tyDenote τ1 -> tyDenote τ2 :=
  match o with
  | Not => negb
  end.

Fixpoint expDenote Δ τ (e : exp Δ τ) (δ : env Δ) : tyDenote τ :=
  match e with
  | Var v => hget δ v
  | Lit l => l
  | Binop o e1 e2 => (binopDenote o) (expDenote e1 δ) (expDenote e2 δ)
  | Unop o e' => (unopDenote o) (expDenote e' δ)
  end.

Fixpoint arg_envDenote ts Δ (δ : env Δ) (ae : arg_env ts Δ) : env ts :=
  match ae with
  | HNil => HNil
  | HCons e ae' => HCons (expDenote e δ) (arg_envDenote δ ae')
  end.

Definition liftVal Δ τ (v : tyDenote τ) : exp Δ τ := Lit v.

Fixpoint envToSub Δ Δ' (δ : env Δ') : Sub (Δ' ++ Δ) Δ :=
  match δ with
  | HNil => idSub
  | HCons e δ' => consSub (liftVal Δ _ e) (envToSub _ δ')
  end.

Fixpoint subDenote Δ Δ' : Sub Δ' Δ -> env Δ -> env Δ' :=
  match Δ' with
  | nil => fun _ _ => HNil
  | cons _ _ => fun s δ => HCons (expDenote (hdSub s) δ) (subDenote (tlSub s) δ)
  end.

Fixpoint renDenote Δ Δ' : Ren Δ' Δ -> env Δ -> env Δ' :=
  match Δ' with
  | nil => fun _ _ => HNil
  | cons _ _ => fun s δ => HCons (hget δ (hdRen s)) (renDenote (tlRen s) δ)
  end.

Lemma renDenoteComm :
  forall Δ τ (e : exp Δ τ) Δ' (r : Ren Δ Δ'),
  forall δ, expDenote e (renDenote r δ) = expDenote (RTmExp r e) δ.
  intros.
  induction e; auto.

  simpl.
  induction m; auto.
  simpl.
  apply (IHm (tlRen r)).

  Rewrites x.
  Rewrites x.
Qed.

Lemma subDenoteComm :
  forall Δ τ (e : exp Δ τ) Δ' (s : Sub Δ Δ'),
  forall δ, expDenote e (subDenote s δ) = expDenote (SubExp s e) δ.
Proof.
  intros.
  induction e; auto.

  simpl.
  induction m; auto.
  simpl.
  apply (IHm (tlSub s)).

  Rewrites x.
  Rewrites x.
Qed.

Lemma subArgEnvDenoteComm :
  forall ts Δ (ae : arg_env ts Δ) Δ' (s : Sub Δ Δ'),
  forall δ, arg_envDenote (subDenote s δ) ae = arg_envDenote δ (SubstArgEnv s ae).
Proof.
  intros.
  induction ae; auto.
  Rewrites subDenoteComm.
Qed.

Lemma hdTailId : forall Δ Δ' τ (s : Sub (τ :: Δ) Δ'),
    consSub (hdSub s) (tlSub s) = s.
Proof.
  intros.
  ExtVar.
Qed.

Lemma consTlSub : forall τ Δ Δ' (s : Sub Δ Δ') (e : exp Δ' τ),
  tlSub (consSub e s) = s.
Proof.
  intros.
  ExtVar.
Qed.


Lemma shiftDenotExp: forall Δ τ (e : exp Δ τ) τ' (δ : env (τ' :: Δ)),
    expDenote (ShTmExp τ' e) δ = expDenote e (htail δ).
  induction e; intros; auto.
  + simpl. dependent destruction δ; auto.
  + simpl. rewrite IHe1. rewrite IHe2. reflexivity.
  + simpl. rewrite IHe. reflexivity.
Qed.

Lemma shiftDenotSub: forall Δ Δ' τ (s : Sub Δ Δ') (δ : env (τ :: Δ')),
    subDenote (tlSub (STmL s)) δ = (subDenote s (htail δ)).
  induction Δ; intros; simpl.
  + reflexivity.
  + unfold hdSub, tlSub in *. simpl in *.
    rewrite IHΔ. rewrite shiftDenotExp. reflexivity.
Qed.

Lemma shiftDenotSub_alt : forall Δ' Δ τ (s:Sub Δ' Δ)(δ : env (τ :: Δ)),
    subDenote (STmL s) δ = HCons (hget δ HFirst) (subDenote s (htail δ)).
  intros. simpl. f_equal. apply shiftDenotSub.
Qed.

Lemma addEventComm : forall e Δ Δ' (s : Sub Δ Δ') (δ : env Δ'),
    subDenote (STmL4 s) (addEvent e δ) = addEvent e (subDenote s δ).
  intros.
  destruct e. unfold addEvent. unfold STmL4.
  remember ((STmL (STmL (STmL s)))) as s'.
  remember ((HCons (t:tyDenote Agent) (HCons (t0:tyDenote Agent) (HCons (t1 : tyDenote Resource) (HCons (t2:tyDenote Timestamp) δ))))) as δ'.
  transitivity (HCons (hget δ' HFirst) (subDenote s' (htail δ'))).
  exact (shiftDenotSub_alt s' δ').
  subst δ'; subst.
  f_equal.
  remember (STmL (STmL s)) as s'.
  remember ((HCons (t0:tyDenote Agent) (HCons (t1 : tyDenote Resource) (HCons (t2: tyDenote Timestamp) δ)))) as δ'.
  replace (htail (HCons (t:tyDenote Agent) δ')) with δ'
    by (subst δ'; reflexivity).
  rewrite (shiftDenotSub_alt s' δ').
  subst s' δ'.
  f_equal.
  rewrite shiftDenotSub_alt. simpl. f_equal.
  f_equal.
  rewrite shiftDenotSub. simpl. reflexivity.
Qed.

Hint Resolve hget_dec tyDenote_dec_eq : csl.

Lemma expDenote_dec : forall Δ τ (e : exp Δ τ) (δ : env Δ) (v : tyDenote τ), {expDenote e δ = v} + {expDenote e δ <> v}.
Proof.
  intros.
  dependent induction e; intros; try (simpl in *;
                                      repeat decide equality); auto with csl.
Qed.

Lemma env_inhabited : forall Δ, inhabited (env Δ).
Proof.
  intros.
  induction Δ.
  constructor.
  apply HNil.
  destruct IHΔ.
  destruct a; constructor; constructor; auto; constructor.
Qed.
