Require Import String List Program FunctionalExtensionality EqNat.
Require Import CSL.HList CSL.Definitions.

Import ListNotations.

Set Implicit Arguments.

Inductive Comparable : ty -> Set :=
| ResourceCmp : Comparable Resource
| ScaleCmp : Comparable Scale
| TimestampCmp : Comparable Timestamp.

Inductive Addable : ty -> Set :=
| ResourceAdd : Addable Resource
| ScaleAdd : Addable Scale.

Inductive Equatable : ty -> Set :=
| AgentEq : Equatable Agent
| ResourceEq: Equatable Resource
| TimestampEq: Equatable Timestamp
| BoolEq: Equatable Bool
| ScaleEq: Equatable Scale.

Definition tyToEquatable (τ: ty) : Equatable τ :=
  match τ with
  | Agent => AgentEq
  | Resource => ResourceEq
  | Timestamp => TimestampEq
  | Bool => BoolEq
  | Scale => ScaleEq
  end.

Inductive binop : ty -> ty -> ty -> Set :=
| And : binop Bool Bool Bool
| Or : binop Bool Bool Bool
| Mult : binop Scale Resource Resource
| Plus : forall τ, Addable τ -> binop τ τ τ
| Eq : forall τ, Equatable τ -> binop τ τ Bool
| Le : forall τ, Comparable τ -> binop τ τ Bool.

Inductive unop : ty -> ty -> Set :=
| Not : unop Bool Bool.

Inductive exp Δ : ty -> Type :=
(* De Bruijn indexed variables under transfers *)
| Var   : forall τ , member τ Δ -> exp Δ τ
| Lit   : forall τ , tyDenote τ -> exp Δ τ
| Binop : forall t1 t2 t3, binop t1 t2 t3 -> exp Δ t1 -> exp Δ t2 -> exp Δ t3
| Unop  : forall t1 t2, unop t1 t2 -> exp Δ t1 -> exp Δ t2.

Arguments Var [Δ τ].
Arguments Lit [Δ τ].
Arguments Binop [Δ t1 t2 t3].
Arguments Unop [Δ t1 t2].

(* HList of expressions with different types for arguments *)
Definition arg_env ts Δ := hlist (exp Δ) ts.

Inductive contract Γ Δ : Type :=
| success    : contract Γ Δ
| failure    : contract Γ Δ
| tapp       : forall ts, arg_env ts Δ -> member ts Γ -> contract Γ Δ
| transfer   : exp (new_var Δ) Bool -> contract Γ (new_var Δ) -> contract Γ Δ
| alternate  : contract Γ Δ -> contract Γ Δ -> contract Γ Δ
| sequence   : contract Γ Δ -> contract Γ Δ -> contract Γ Δ
| concurrent : contract Γ Δ -> contract Γ Δ -> contract Γ Δ.

Definition template_env Γ := hlist (contract Γ) Γ.

Arguments transfer [Γ Δ].
Arguments alternate [Γ Δ].
Arguments sequence [Γ Δ].
Arguments concurrent [Γ Δ].

Notation "A ;; B" := (sequence A B) (at level 80, right associativity).
Notation "A :+: B" := (alternate A B) (at level 79, left associativity).
Notation "A :||: B" := (concurrent A B) (at level 78, left associativity).

Inductive TopLevel {Γ Δ}: template_env Γ -> contract Γ Δ -> Type :=
| LetRec: forall (D : template_env Γ) (c : contract Γ Δ), TopLevel D c.

(* Substitutions ala Typed Term Representation by Benton et al *)
Definition Ren Δ Δ' := forall τ, @member ty τ Δ -> member τ Δ'.

Definition tlRen {Δ Δ' τ} (r : Ren (τ :: Δ) Δ') : Ren Δ Δ' :=
  fun τ' v => r τ' (HNext v).

Definition hdRen {Δ Δ' τ} (r : Ren (τ :: Δ) Δ') : member τ Δ' :=
  r τ HFirst.

Program Definition RTmL {Δ Δ' τ} (r : Ren Δ Δ') : Ren (τ :: Δ) (τ :: Δ') :=
  fun τ' v =>
    match v with
    | HFirst => @HFirst _ _ Δ'
    | HNext v' => HNext (r _ v')
    end.


Fixpoint RTmExp Δ Δ' τ (r : Ren Δ Δ') (e : exp Δ τ) :=
  match e with
  | Var v => Var (r _ v)
  | Lit l => Lit l
  | Binop o e1 e2 => Binop o (RTmExp r e1) (RTmExp r e2)
  | Unop o e' => Unop o (RTmExp r e')
  end.

Definition ShTmExp Δ τ τ' : exp Δ τ -> exp (τ' :: Δ) τ :=
  RTmExp (fun _ v => HNext v).

Definition Sub Δ Δ' := forall τ, member τ Δ -> exp Δ' τ.

Definition idSub {Δ} : Sub Δ Δ := @Var Δ.

Program Definition consSub {Δ Δ' τ} (e : exp Δ' τ) (s : Sub Δ Δ'): Sub (τ :: Δ) Δ' :=
  fun τ' (v : member τ' (τ :: Δ)) =>
        match v with
        | HFirst => e
        | HNext v' => s _ v'
        end.

Definition tlSub {Δ Δ' τ} (s : Sub (τ :: Δ) Δ') : Sub Δ Δ' :=
  fun τ' v => s τ' (HNext v).

Definition hdSub {Δ Δ' τ} (s : Sub (τ :: Δ) Δ') : exp Δ' τ :=
  s τ HFirst.

Fixpoint SubExp Δ Δ' τ (s : Sub Δ Δ') (e : exp Δ τ) : exp Δ' τ :=
  match e with
  | Var v => s _ v
  | Lit l => Lit l
  | Binop o e1 e2 => Binop o (SubExp s e1) (SubExp s e2)
  | Unop o e' => Unop o (SubExp s e')
  end.

Program Definition STmL {Δ Δ'} τ (s : Sub Δ Δ') : Sub (τ :: Δ) (τ :: Δ') :=
  fun τ' v =>
    match v with
    | HFirst => Var HFirst
    | HNext v' => ShTmExp τ (s _ v')
    end.


Definition STmL4 {Δ Δ'} (s : Sub Δ Δ') : Sub (new_var Δ) (new_var Δ') :=
  STmL (STmL (STmL (STmL s))).


Fixpoint SubstArgEnv ts Δ Δ' (s : Sub Δ Δ') (ae : arg_env ts Δ) : arg_env ts Δ' :=
  match ae with
  | HNil => HNil
  | HCons e ae' => HCons (SubExp s e) (SubstArgEnv s ae')
  end.

Fixpoint SubstContract Γ Δ Δ' (s : Sub Δ Δ') (c : contract Γ Δ) : contract Γ Δ' :=
  match c with
  | success _ _ => success _ _
  | failure _ _ => failure _ _
  | tapp ae f => tapp (SubstArgEnv s ae) f
  | transfer p c => transfer (SubExp (STmL4 s) p) (SubstContract (STmL4 s) c)
  | alternate c1 c2 => alternate (SubstContract s c1) (SubstContract s c2)
  | sequence c1 c2 => sequence (SubstContract s c1) (SubstContract s c2)
  | concurrent c1 c2 => concurrent (SubstContract s c1) (SubstContract s c2)
  end.

Notation "{| e ; .. ; f |}" := (consSub e .. (consSub f idSub) ..).

Ltac Rewrites E :=
  (intros; simpl; try rewrite E;
   repeat (match goal with
           | [ H : context[_ = _] |- _ ] => rewrite H
           end);
   auto).

Ltac ExtVar :=
  match goal with
    [ |- ?X = ?Y ] =>
    (apply (@functional_extensionality_dep _ _ X Y);
     let t := fresh "t"
     in intro t;
        apply functional_extensionality;
        let v := fresh "v"
        in intro v;
           dependent destruction v; auto)
      end.

Lemma LiftIdSub : forall Δ τ, STmL (@idSub Δ) = @idSub (τ :: Δ).
Proof. intros. ExtVar. Qed.

Lemma ActIdSub : forall Δ τ (e : exp Δ τ),
    SubExp idSub e = e.
Proof. induction e; Rewrites LiftIdSub. Qed.

Definition RcR {Δ Δ' Δ''} (r : Ren Δ' Δ'') (r' : Ren Δ Δ') : Ren Δ Δ'' :=
  fun τ v => r τ (r' τ v).
Definition ScR {Δ Δ' Δ''} (s : Sub Δ' Δ'') (r : Ren Δ Δ') : Sub Δ Δ'' :=
  fun τ v => s τ (r τ v).
Definition RcS {Δ Δ' Δ''} (r : Ren Δ' Δ'') (s : Sub Δ Δ') : Sub Δ Δ'' :=
  fun τ v => RTmExp r (s τ v).
Definition ScS {Δ Δ' Δ''} (s : Sub Δ' Δ'') (s' : Sub Δ Δ') : Sub Δ Δ'' :=
  fun τ v => SubExp s (s' τ v).

Lemma LiftRcR : forall Δ Δ' Δ'' τ (r : Ren Δ' Δ'') (r' : Ren Δ Δ'),
    RTmL (τ := τ) (RcR r r') = RcR (RTmL r) (RTmL r').
Proof. intros. ExtVar. Qed.

Lemma ActRcR : forall Δ τ (e : exp Δ τ) Δ' Δ'' (r : Ren Δ' Δ'') (r' : Ren Δ Δ'),
    RTmExp (RcR r r') e = RTmExp r (RTmExp r' e).
Proof. induction e; Rewrites LiftRcR. Qed.

Lemma LiftScR : forall Δ Δ' Δ'' τ (s : Sub Δ' Δ'') (r : Ren Δ Δ'),
    STmL (τ := τ) (ScR s r) = ScR (STmL s) (RTmL r).
Proof. intros. ExtVar. Qed.

Lemma ActScR : forall Δ τ (e : exp Δ τ) Δ' Δ'' (s : Sub Δ' Δ'') (r : Ren Δ Δ'),
    SubExp (ScR s r) e = SubExp s (RTmExp r e).
Proof. induction e; Rewrites LiftRcR. Qed.

Lemma LiftRcS : forall Δ Δ' Δ'' τ (r : Ren Δ' Δ'') (s : Sub Δ Δ'),
    STmL (τ := τ) (RcS r s) = RcS (RTmL r) (STmL s).
Proof. intros. ExtVar. unfold RcS. simpl.
       unfold ShTmExp. rewrite <- 2 ActRcR. auto. Qed.

Lemma ActRcS : forall Δ τ (e : exp Δ τ) Δ' Δ'' (r : Ren Δ' Δ'') (s : Sub Δ Δ'),
    SubExp (RcS r s) e = RTmExp r (SubExp s e).
Proof. induction e; Rewrites LiftRcS. Qed.

Lemma LiftScS : forall Δ Δ' Δ'' τ (s : Sub Δ' Δ'') (s' : Sub Δ Δ'),
    STmL (τ := τ) (ScS s s') = ScS (STmL s) (STmL s').
Proof. intros. ExtVar. unfold ScS. simpl.
       unfold ShTmExp. rewrite <- ActRcS. rewrite <- ActScR. auto. Qed.

Lemma ActScS : forall Δ τ (e : exp Δ τ) Δ' Δ'' (s : Sub Δ' Δ'') (s' : Sub Δ Δ'),
    SubExp (ScS s s') e = SubExp s (SubExp s' e).
Proof. induction e; Rewrites LiftScS. Qed.
